/* # skkinput (Simple Kana-Kanji Input)
 * SkkInput.c --- Making OverTheSpotWinWidget.
 * This file is part of skkinput.
 * Copyright (C) 1997
 * Takashi SAKAMOTO (sakamoto@yajima.kuis.kyoto-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "OverWinP.h"
#include "config.h"
#include "skkkey.h"
#include "otwexfunc.h"
#include "SubwindowShell.h"
#include "Canvas.h"
#include "WmcloseShell.h"

#ifndef XtNgeometry
#define XtNgeometry		"geometry"
#define XtCGeometry		"Geometry"
#endif

#define offset(field) XtOffsetOf( OverthespotWinRec, overthespotWin.field)
#define goffset(field) XtOffsetOf( WidgetRec, core.field )

static XtResource OTSW_resources[] = {
  /* Ȥ餢꥽*/
  { XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
    goffset(width), XtRImmediate, (XtPointer) 400},
  { XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
    goffset(height), XtRImmediate, (XtPointer) 400},
  /* ߥ˥Хåե礭ꤹġǥեȤΤ͡١ꥵ *
   * Ф褦ˤ褦 */
  { XtNmwidth, XtCMwidth, XtRDimension, sizeof(Dimension),
    offset(minibuf_width), XtRImmediate, (XtPointer)80 },
  /* ʬȤǺä꥽*/
  { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
    offset(puppixel), XtRString, XtDefaultForeground },
  /* Minibuffer ȿžɽ뤫ɤȤäˤʤ롣*/
  { XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
    offset(reverse_video), XtRImmediate, (XtPointer) FALSE},
  /* Overthespot Ѥեȵڤӥ޻եȡ*/
  { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    offset(fs_default_roman), XtRString, DEFAULT_ROMANFONT },
  { XtNkanjiFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    offset(fs_default_kanji), XtRString, DEFAULT_KANJIFONT },
  /* Minibuffer Window 褹ΤѤڤӥ޻եȡ*/
  { XtNmfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    offset(fs_minibuf_roman), XtRString, DEFAULT_ROMANFONT },
  { XtNmkanjiFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    offset(fs_minibuf_kanji), XtRString, DEFAULT_KANJIFONT },
  /* Minibuffer Window 礭եȤν˺뤫ɤ*/
  { XtNchangeMinibufferFont, XtCChangeMinibufferFont,
    XtRBoolean, sizeof ( Boolean ),
    offset(change_minibuffer_font), XtRImmediate, (XtPointer)TRUE },
  /* skkinput 뤬ĤѤ callback */
  { XtNendNotify, XtCCallback, XtRCallback, sizeof(caddr_t), 
    offset(endcallback), XtRCallback, (caddr_t)NULL },
  /* skkinput  client ʸѤ callback */
  { XtNfixNotify, XtCCallback, XtRCallback, sizeof(caddr_t), 
    offset(fixcallback), XtRCallback, (caddr_t)NULL },
  /* key event  client ֤Ѥ callback */
  { XtNkeybackNotify, XtCCallback, XtRCallback, sizeof(caddr_t), 
    offset(keybackcallback), XtRCallback, (caddr_t)NULL },
  /* egg ߴ j-newline ݤ*/
  { XtNeggLikeNewline, XtCEggLikeNewline, XtRImmediate, sizeof (int),
    offset(egg_like_newline), XtRImmediate, (XtPointer) FALSE},
  /* chat-adapter-mode ݤ*/
  { XtNchatAdapter, XtCChatAdapter, XtRImmediate, sizeof (int),
    offset(chat_adapter), XtRImmediate, (XtPointer) FALSE},
  { XtNjisyoDirty, XtCJisyoDirty, XtRImmediate, sizeof (int),
    offset(jisyo_dirty), XtRImmediate, (XtPointer) FALSE },
  { XtNconversionAttribute, XtCConversionAttribute, XtRImmediate, 
    sizeof( struct ConvAttrsMesg * ),
    offset( camsg ), XtRImmediate, ( XtPointer )NULL },
  /* ߥ˥Хåեξõ롣*/
  { XtNclearMinibuffer, XtCClearMinibuffer, XtRImmediate, sizeof( int ),
    offset( clearMinibuffer ), XtRImmediate, (XtPointer)FALSE },
  { XtNclientWindow, XtCClientWindow, XtRImmediate, sizeof( Window ),
    offset( client_window ), XtRImmediate, (XtPointer)None },
  { XtNsetFocus, XtCSetFocus, XtRImmediate, sizeof( int ),
    offset( conversion_set_focus ), XtRImmediate, (XtPointer)FALSE },
  { XtNunsetFocus, XtCUnsetFocus, XtRImmediate, sizeof( int ),
    offset( conversion_unset_focus ), XtRImmediate, (XtPointer)FALSE  },
  /* Subwindow ե֤ɥꤹ롣*/
  { XtNprobeWindow, XtCProbeWindow, XtRImmediate, sizeof( Window ),
    offset( probe_window ), XtRImmediate, ( XtPointer )None },
} ;

#undef offset
#undef goffset

/*
 * ץȥ
 */
/*
 * ΥåȤƤǤȤʤؿ
 */
static void OTSW_Initialize
( Widget greq, Widget gnew, ArgList args, Cardinal *num_args ) ;
static void OTSW_Realize
( Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs ) ;
static void OTSW_Redisplay
( Widget gw, XEvent *event, Region region ) ;
static void OTSW_Destroy( Widget gw ) ;
static Boolean OTSW_SetValues
( Widget current, Widget request, Widget new,
  ArgList args, Cardinal *num_args ) ;
static void KeyDownEventHandler
( Widget w, XEvent *event, String *params, Cardinal *num_params ) ;
static void FocusEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params ) ;

static void OTSW_CreateAllGCs( Widget gw ) ;
static void OTSW_CreateMinibufferGCs( Widget gw ) ;
static void OTSW_ReleaseAllGCs( Widget gw ) ;
static void OTSW_ReleaseMinibufferGCs( Widget gw ) ;

static int OTSW_ChangeAttributes
( Widget gw, unsigned long valuemask, struct ConvAttrs *values ) ;
static void OTSW_resetupModeWindow( Widget gw ) ;
static void OTSW_ReconfigureColor
( Widget gw, unsigned long valuemask ) ;

static void OTSW_ConfigureMinibuffer( Widget gw ) ;

static int do_function_by_reading_character
( Widget gw, struct SKKInputNode *node ) ;
static void OTSW_RedrawScreen( Widget gw ) ;

static void OTSW_ConfigureWindowsScreenIsSmall
( Widget gw, int endx, int endy, 
  unsigned char *linetop,     int linetop_pos,
  unsigned char *midline_top, int midline_pos ) ;
static int OTSW_ReconfigureFocusPosition
( Widget gw ) ;

static int focusCheck( Widget gw ) ;
/*
 * callbacks
 */
static void Topline_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void Midline_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void Botline_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void Minibuffer_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void Modeshell_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void Canvas_DestroyCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void OTSW_wmMessageCloseCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void OTSW_wmShellDestroyCallback
( Widget gw, caddr_t client, caddr_t caller ) ;

/*
 * Window ˴ؤؿΥץȥ 
 */
extern void skkwin_jputstring
( Display *disp, Window win, GC roman_gc, GC kanji_gc,
  int x, int y, char *text, int roman_width, int kanji_width ) ;
extern void skkwin_jputchar
( Display *disp, Window win, GC roman_gc, GC kanji_gc,
  int x, int y, int chara, int flag ) ;
extern void skkinputDrawCursor
( Display *disp, Window win, GC roman_ngc, GC roman_rgc, 
  int x, int y, int font_ascent, int width, int height,
  int is_focus ) ;

/* MyDispatch.c */
extern int add_myeventhandler
( Display *disp, Window src_win, Window dest_win,
  int event_type, long event_mask ) ;
extern int remove_allmyeventhandler
( Display *disp, Window src_win ) ;
extern GC my_CreateGC
( Widget gw, Pixel foreground, Pixel background, XFontStruct *font ) ;
extern void getFontInfo
( XFontStruct *fs_kanji, XFontStruct *fs_roman,
  int *ret_kanji_width, int *ret_roman_width,
  int *ret_font_height, int *ret_font_ascent ) ;
extern void OTSW_CalcTopPositionOfBuffer
( unsigned char *linetop, int linetop_pos, int cur_pos,
  int client_width, int roman_width, int kanji_width,
  unsigned char **ret_toptext, int *ret_toppos ) ; 
extern void OTSW_FullDrawTopBufferTopLine
( Display *disp, Window win,
  GC roman_ngc, GC kanji_ngc, GC roman_rgc, GC kanji_rgc,
  unsigned char *text, int text_pos,
  int roman_width, int kanji_width, int font_ascent, int font_height,
  int width, int bsflag, int cur_pos, int cur_exist, int is_focus ) ;
extern void XChangeWMNormalHints
( Display *disp, Window win, XSizeHints *hints ) ;
extern void XGetRectangleOfWindow
( Display *disp, Window win, XRectangle *area ) ;

/* MyError.c */
extern void ignore_xerror( void ) ;
extern void regard_xerror( void ) ;

/*
 * Хѿ
 */
static XtActionsRec overthespotWinActionsTable [] = {
  { "KeyDownEventHandler",	   KeyDownEventHandler },
  { "FocusInEventHandler",	   FocusEventHandler },
  { "FocusOutEventHandler",	   FocusEventHandler },
};

static char defaultOverthespotWinTranslations[] =  
"<Key>:                   KeyDownEventHandler()\n\
 <KeyRelease>:            KeyDownEventHandler()\n\
 <FocusIn>:               FocusInEventHandler()\n\
 <FocusOut>:              FocusOutEventHandler()\n" ;

OverthespotWinClassRec overthespotWinClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Overthespot",
    /* size			*/	sizeof( OverthespotWinRec ),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	OTSW_Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	OTSW_Realize,
    /* actions			*/	overthespotWinActionsTable,
    /* num_actions		*/	XtNumber( overthespotWinActionsTable ),
    /* resources		*/	OTSW_resources,
    /* num_resources		*/	XtNumber( OTSW_resources ),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	OTSW_Destroy,
    /* resize			*/	NULL,
    /* expose			*/	OTSW_Redisplay,
    /* set_values		*/	OTSW_SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	defaultOverthespotWinTranslations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    }
};

WidgetClass overthespotWinWidgetClass =
( WidgetClass )&overthespotWinClassRec ;

static void OTSW_Initialize 
( Widget greq, Widget gnew, ArgList args, Cardinal *num_args )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gnew ;
  struct SKKInputNode *node ;
  
  /* ե٥åɽΤ GC 롣*/
  w->overthespotWin.background = w->core.background_pixel ;
  w->overthespotWin.foreground = w->overthespotWin.puppixel ;
  w->overthespotWin.colormap   = w->core.colormap ;

  w->overthespotWin.oforeground = w->overthespotWin.foreground ;
  w->overthespotWin.obackground = w->overthespotWin.background ;

  /* ǥեȤΥեȤˤƤ*/
  w->overthespotWin.fs_roman = w->overthespotWin.fs_default_roman ;
  w->overthespotWin.fs_kanji = w->overthespotWin.fs_default_kanji ;

  /* եȤξȴФƤ*/
  getFontInfo
    ( w->overthespotWin.fs_kanji, w->overthespotWin.fs_roman,
      &w->overthespotWin.kanji_width, &w->overthespotWin.roman_width,
      &w->overthespotWin.font_height, &w->overthespotWin.font_ascent ) ;
  /* ߥ˥ХåեΥեȤξȴФƤ*/
  getFontInfo
    ( w->overthespotWin.fs_minibuf_kanji, w->overthespotWin.fs_minibuf_roman,
      &w->overthespotWin.mkanji_width, &w->overthespotWin.mroman_width,
      &w->overthespotWin.mfont_height, &w->overthespotWin.mfont_ascent ) ;
  
  OTSW_CreateAllGCs( gnew ) ;
  OTSW_CreateMinibufferGCs( gnew ) ;

  /* ¾ν*/
  w->overthespotWin.cutbuffer[ 0 ]  = '\0' ;
  w->overthespotWin.keybuf_use    = 0 ;
  w->overthespotWin.keybuf[ 0 ]   = '\0' ;

  /* ҥȥϥꥢƤɸνɬפǤĤޤϰ */
  /* ⵯ줿ȤʤƤʤΤǴǤ롣*/

  /* ԽѤΥХåեγݡ*/
  node = skkinput_AllocateMinibuffer() ;
  node->cur_exist = True ;
  w->overthespotWin.topbuffer  = node ;
  w->overthespotWin.lastbuffer = node ;

  w->overthespotWin.jisyo_dirty = 0 ;

  /* ޤFocus ϺǽϳƤȻפ衣*/
  w->overthespotWin.is_focus = False ;

  /* Over-the-spot Window ȤƵǽ뤿ɬפѿν*/
  w->overthespotWin.focus_window = None ;
  w->overthespotWin.input_focus_window = None ;
  w->overthespotWin.spot_x = w->overthespotWin.spot_y = 0 ;
  w->overthespotWin.focus_offset_x =
    w->overthespotWin.focus_offset_y = 0 ;
  w->overthespotWin.client_x =
    w->overthespotWin.client_y = 0 ;
  w->overthespotWin.client_area.x = 
    w->overthespotWin.client_area.y = 0 ;
  w->overthespotWin.client_area.height = 
    w->overthespotWin.client_area.width = 0 ;
  w->overthespotWin.status_area.x = 
    w->overthespotWin.status_area.y = 0 ;
  w->overthespotWin.status_area.height = 
    w->overthespotWin.status_area.width = 0 ;

  /* ХåեȤɽ뤿˻ȤɥõƤ*/
  w->overthespotWin.topline_probe = False ;
  w->overthespotWin.midline_probe = False ;
  w->overthespotWin.botline_probe = False ;
  w->overthespotWin.minibuffer_probe = False ;
  w->overthespotWin.modeshell_probe = False ;

  w->overthespotWin.attribute_mask = 0L ;
  return ;
}

static void OTSW_ReleaseAllGCs( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;

  XtReleaseGC( gw, w->overthespotWin.roman_ngc ) ;
  XtReleaseGC( gw, w->overthespotWin.roman_rgc ) ;
  XtReleaseGC( gw, w->overthespotWin.kanji_ngc ) ;
  XtReleaseGC( gw, w->overthespotWin.kanji_rgc ) ;
  return ;
}

static void OTSW_ReleaseMinibufferGCs( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;

  XtReleaseGC( gw, w->overthespotWin.mroman_ngc ) ;
  XtReleaseGC( gw, w->overthespotWin.mroman_rgc ) ;
  XtReleaseGC( gw, w->overthespotWin.mkanji_ngc ) ;
  XtReleaseGC( gw, w->overthespotWin.mkanji_rgc ) ;
  return ;
}

static void OTSW_CreateAllGCs( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;

  w->overthespotWin.roman_ngc = 
    my_CreateGC
      ( gw, w->overthespotWin.oforeground, w->overthespotWin.obackground,
       w->overthespotWin.fs_roman ) ;
  w->overthespotWin.kanji_ngc = 
    my_CreateGC
      ( gw, w->overthespotWin.oforeground, w->overthespotWin.obackground,
       w->overthespotWin.fs_kanji ) ;
  w->overthespotWin.roman_rgc = 
    my_CreateGC
      ( gw, w->overthespotWin.obackground, w->overthespotWin.oforeground, 
       w->overthespotWin.fs_roman ) ;
  w->overthespotWin.kanji_rgc = 
    my_CreateGC
      ( gw, w->overthespotWin.obackground, w->overthespotWin.oforeground, 
       w->overthespotWin.fs_kanji ) ;
  return ;
}

static void OTSW_CreateMinibufferGCs( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  Pixel foreground, background ;

  if( w->overthespotWin.change_minibuffer_font ){
    foreground = w->overthespotWin.oforeground ;
    foreground = w->overthespotWin.obackground ;
  } else {
    foreground = w->overthespotWin.foreground ;
    background = w->overthespotWin.background ;
  }
  w->overthespotWin.mroman_ngc = 
    my_CreateGC
    ( gw, foreground, background,
      w->overthespotWin.fs_minibuf_roman ) ;
  w->overthespotWin.mkanji_ngc = 
    my_CreateGC
    ( gw, foreground, background,
      w->overthespotWin.fs_minibuf_kanji ) ;
  w->overthespotWin.mroman_rgc = 
    my_CreateGC
    ( gw, background, foreground, 
      w->overthespotWin.fs_minibuf_roman ) ;
  w->overthespotWin.mkanji_rgc = 
    my_CreateGC
    ( gw, background, foreground, 
      w->overthespotWin.fs_minibuf_kanji ) ;
  return ;
}

static void OTSW_CreateCanvasOfTopBuffer
( Widget gw, Window focus_window, Window probe_window,
  Widget *popup_gw, Widget *canvas_gw,
  int *probe, XtCallbackProc callbackproc )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  Widget line_popup, line_canvas ;

  line_popup = XtVaCreatePopupShell
    ( "topbuffer", subwindowShellWidgetClass, gw, XtNinput, TRUE, 
      XtNmappedWhenManaged, TRUE, XtNwidth, 1, XtNheight, 1, 
      XtNcolormap, w->overthespotWin.colormap,
      XtNbackground, w->overthespotWin.obackground, 
      XtNparentWindow, focus_window,
      XtNprobeWindow, probe_window, NULL ) ;
  /* βˤĤɽѤ widget 롣shellwidget ľ *
   * ܽ񤤤ƤϤޤäȤΤǡġ*/
  line_canvas = XtVaCreateManagedWidget
    ( "canvas", canvasWidgetClass, line_popup, NULL ) ;
  /* Ѥ륳ХåϿ롣*/
  XtAddCallback
    ( line_canvas, XtNredrawNotify, callbackproc, gw ) ;
  XtAddCallback
    ( line_canvas, XtNdestroyNotify, 
      ( XtCallbackProc )Canvas_DestroyCallback, gw ) ;
  /* åȤβ롣*/
  XtRealizeWidget( line_popup ) ;
  XtRealizeWidget( line_canvas ) ;
  *probe = False ;
  *popup_gw  = line_popup ;
  *canvas_gw = line_canvas ;
  return ;
}

/*
 * ߥ˥ХåեβΤ˸ƤӽФؿ
 */
static void OTSW_RealizeMinibuffer( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  Widget minibuffer_popup, minibuffer_canvas ;
  XSizeHints sizehints ;
  unsigned int winwidth, winheight, roman_width, roman_height ;
  Pixel foreground, background ;

  if( w->overthespotWin.change_minibuffer_font ){
    winwidth  = w->overthespotWin.roman_width *
      w->overthespotWin.minibuf_width ;
    winheight = w->overthespotWin.font_height ;
    roman_width  = w->overthespotWin.roman_width ;
    roman_height = w->overthespotWin.font_height ;
  } else {
    winwidth  = w->overthespotWin.mroman_width * 
      w->overthespotWin.minibuf_width ;
    winheight = w->overthespotWin.mfont_height ;
    roman_width  = w->overthespotWin.mroman_width ;
    roman_height = w->overthespotWin.mfont_height ;
  }
  if( w->overthespotWin.change_minibuffer_font ){
    foreground = w->overthespotWin.oforeground ;
    background = w->overthespotWin.obackground ;
  } else {
    foreground = w->overthespotWin.foreground ;
    background = w->overthespotWin.background ;
  }
  minibuffer_popup = XtVaCreatePopupShell
    ( "minibuffer", wmcloseShellWidgetClass, gw, XtNinput, TRUE, 
      XtNmappedWhenManaged, TRUE, 
      XtNwidth, winwidth, XtNheight, winheight, 
      XtNcolormap, w->overthespotWin.colormap,
      XtNreverseVideo, w->overthespotWin.reverse_video,
      XtNbackground, background,
      XtNforeground, foreground, NULL ) ;
  XtAddCallback
    ( minibuffer_popup, XtNwmcloseNotify, 
      ( XtCallbackProc )OTSW_wmMessageCloseCallback, gw ) ;
  XtAddCallback
    ( minibuffer_popup, XtNdestroyNotify, 
      ( XtCallbackProc )OTSW_wmShellDestroyCallback, gw ) ;

  /* βˤĤɽѤ widget 롣shellwidget ľ *
   * ܽ񤤤ƤϤޤäȤΤǡġ*/
  minibuffer_canvas = XtVaCreateManagedWidget
    ( "canvas", canvasWidgetClass, minibuffer_popup,
      XtNreverseVideo, w->overthespotWin.reverse_video,
      XtNbackground, background, 
      XtNforeground, foreground, NULL ) ;
  w->overthespotWin.minibuffer_probe = False ;
  /* 襤٥ȤåäƤ*/
  XtAddCallback
    ( minibuffer_canvas, XtNredrawNotify,
      ( XtCallbackProc )Minibuffer_RedisplayCallback, gw ) ;
  XtAddCallback
    ( minibuffer_canvas, XtNdestroyNotify, 
      ( XtCallbackProc )Canvas_DestroyCallback, gw ) ;

  /* ˥ߥ˥ХåեξϤƤrealize Ƥȴ*/
  w->overthespotWin.minibuffer_popup  = minibuffer_popup ;
  w->overthespotWin.minibuffer_canvas = minibuffer_canvas ;

  /* 괺ΤġXtPopDown ȤפʤΤʤ*/
  XtRealizeWidget( minibuffer_popup ) ;
  XtRealizeWidget( minibuffer_canvas ) ;

  /* Ƥ˥ꥵäƤ롩 */
  (void) XtMakeResizeRequest
    ( ( Widget )minibuffer_popup,
      ( Dimension )winwidth, ( Dimension )winheight,
      &minibuffer_popup->core.width,
      &minibuffer_popup->core.height ) ;

  /* ɥΥꥵ¾ꡣ*/
  sizehints.win_gravity = NorthWestGravity ;
  sizehints.base_width  = minibuffer_popup->core.border_width * 2 ;
  sizehints.base_height = minibuffer_popup->core.border_width * 2 ;
  sizehints.width      = sizehints.base_width + winwidth ;
  sizehints.height     = sizehints.base_height + winheight ;
  sizehints.width_inc  = roman_width ;
  sizehints.height_inc = roman_height ;
  sizehints.min_width  = roman_width * 2 + sizehints.base_width ;
  sizehints.min_height = sizehints.base_height + winheight ;
  sizehints.max_width  = roman_width * 160 + sizehints.base_width ;
  sizehints.max_height = sizehints.base_height + winheight ;
  sizehints.flags      =
    ( PBaseSize | PMinSize | PResizeInc | PMaxSize | USSize | PWinGravity ) ;

  /* Window Manager ˥ҥȤäƤ롣*/
  XSetWMNormalHints
    ( XtDisplay( gw ), XtWindow( minibuffer_popup ), &sizehints ) ;
  XFlush( XtDisplay( gw ) ) ;
  return ;
}

/*
 * Widget β˸ƤФؿ
 */
static void OTSW_Realize
( Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  CoreWidgetClass super =
    ( CoreWidgetClass )XtClass( gw )->core_class.superclass ;

  /* ŬڤʥɥѰդƤ*/
  ( *super->core_class.realize )( gw, valueMask, attrs ) ;

  if( w->overthespotWin.client_window == None ){
    XtDestroyWidget( gw ) ;
    return ;
  }
  if( w->overthespotWin.focus_window == None ){
    w->overthespotWin.focus_window = w->overthespotWin.client_window ;
    if( !( w->overthespotWin.attribute_mask & CAClientArea ) )
      XGetRectangleOfWindow
	( XtDisplay( gw ), w->overthespotWin.focus_window, 
	  &w->overthespotWin.client_area ) ;
    OTSW_ReconfigureFocusPosition( gw ) ;
  }
  /* Minibuffer 롣*/
  OTSW_RealizeMinibuffer( gw ) ;

  /* modeshell ɽΤΥåȤ롣*/
  OTSW_CreateCanvasOfTopBuffer
    ( gw, w->overthespotWin.focus_window, 
      w->overthespotWin.probe_window,
      &w->overthespotWin.modeshell_popup,
      &w->overthespotWin.modeshell_canvas,
      &w->overthespotWin.modeshell_probe,
      ( XtCallbackProc )Modeshell_RedisplayCallback ) ;
  XtResizeWidget
    ( w->overthespotWin.modeshell_popup,
      6 * w->overthespotWin.roman_width,
      w->overthespotWin.font_height, gw->core.border_width ) ;

  /* ʸѤΥХ3ɬפˤʤ롣*/
  OTSW_CreateCanvasOfTopBuffer
    ( gw, w->overthespotWin.focus_window, 
      w->overthespotWin.probe_window,
      &w->overthespotWin.topline_popup,
      &w->overthespotWin.topline_canvas,
      &w->overthespotWin.topline_probe,
      ( XtCallbackProc )Topline_RedisplayCallback ) ;
  OTSW_CreateCanvasOfTopBuffer
    ( gw, w->overthespotWin.focus_window, 
      w->overthespotWin.probe_window,
      &w->overthespotWin.midline_popup,
      &w->overthespotWin.midline_canvas,
      &w->overthespotWin.midline_probe,
      ( XtCallbackProc )Midline_RedisplayCallback ) ;
  OTSW_CreateCanvasOfTopBuffer
    ( gw, w->overthespotWin.focus_window, 
      w->overthespotWin.probe_window,
      &w->overthespotWin.botline_popup,
      &w->overthespotWin.botline_canvas,
      &w->overthespotWin.botline_probe,
      ( XtCallbackProc )Botline_RedisplayCallback ) ;
     
  /* եꤹΤʤ顢focus ѲѤ뤹٤Ƥ */
  /* åȤrealizeǤʤФʤʤ*/
  /* ߥ˥Хåեɬ̤򵯤ƤΤǡˤʤġ*/
  XtSetKeyboardFocus( w->overthespotWin.minibuffer_popup, gw ) ;
  /*
   * ΰ֤ʤ⡼ɥƤʤ
   */
  OTSW_resetupModeWindow( gw ) ;
#if 1
  OTSW_ReconfigureColor( gw, CAColor | CAColormap ) ;
#endif
  XSync( XtDisplay( gw ), False ) ;
  return ;
}

/*
 * SIW  XtDestroyWidget 򤫤˸ƤӽФؿ
 *-----
 * SIW νλ·äƤ롣GC Ƥ顢malloc Ƥ
 * Ƥ롣
 */
static void OTSW_Destroy( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  struct SKKInputNode *node, *pNode ;
  /*
   * XFreeFont( XtDisplay( gw ), w->overthespotWin.fs_kanji ) ;
   * XFreeFont( XtDisplay( gw ), w->overthespotWin.fs_roman ) ;
   * ɤ顢Widget ˳ݤƤ櫓ʤߤʤΤǡ
   * ȤϤޤߤ
   */
  ignore_xerror() ;
  /* Destroy ʾ塢ʲΥɥΥ٥Ȥ̵뤹롣*/
  XSelectInput( XtDisplay( gw ), XtWindow( gw ), NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.modeshell_popup ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.modeshell_canvas ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.topline_popup ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.topline_canvas ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.midline_popup ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.midline_canvas ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.botline_popup ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.botline_canvas ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.minibuffer_popup ),
      NoEventMask ) ;
  XSelectInput
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.minibuffer_canvas ),
      NoEventMask ) ;
  /* Event Queue եå夷Ƥ*/
  XSync( XtDisplay( gw ), True ) ;
  regard_xerror() ;

  OTSW_ReleaseAllGCs( gw ) ;
  OTSW_ReleaseMinibufferGCs( gw ) ;
  /*
   * ХåեƲƤ
   */
  node = w->overthespotWin.lastbuffer ;
  while( node != NULL ){
    pNode = node->parentbuffer ;
    free_Minibuffer
      ( gw, &w->overthespotWin.topbuffer,
	&w->overthespotWin.lastbuffer, node ) ;
    node  = pNode ;
  }
  XtCallCallbacks( gw, XtNendNotify, NULL ) ;
  return ;
}

/*
 * ̤褹ؿ
 */
static void OTSW_Redisplay( Widget gw, XEvent *event, Region region )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  /* ϲΥߥ󥰤Ǵְä Widget κ褬Ƥ롣*/
  if( w->overthespotWin.topbuffer == NULL || w->core.being_destroyed ){
#if 1
    fprintf( stderr, "Illegal Widget Exposure Event.\n" ) ;
#endif
    return ;
  }
  OTSW_RedrawScreen( gw ) ;
/*  XFlush( XtDisplay( gw ) ) ;*/
  return ;
}

/*
 * 䤳ȥ饤ȤΥå򸫤ơåȤ
 * ꤹؿ
 *----
 * 饤ȥɥΰưΤ饤ȥɥ
 *  lower ꡢ raise ȡɤ褦̵褦ʵ
 * ġɤˤ臘ʤͤʤȤʤΤɤġ
 */
static int OTSW_ReconfigureWindows( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;

  if( w->overthespotWin.minibuffer_probe ){
    XRaiseWindow
      ( XtDisplay( gw ),
	XtWindow( w->overthespotWin.minibuffer_popup ) ) ;
  }
  /* Spot Location ѲˤʸɽƤ륭Хư */
  /* ʤФʤʤν˺ä C-h ѴưԽ */
  /* ʸõ˥ɥĤ뤳Ȥˤʤ롣*/
  OTSW_RedrawScreen( gw ) ;
  w->overthespotWin.camsg = NULL ;
  return True ;
}

static void OTSW_ResizeMinibufferAndModeshell( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  XSizeHints sizehints ;

  /* modeshell ξϤޤǲȤǤ礦͡*/
  XtResizeWidget
    ( w->overthespotWin.modeshell_popup,
      6 * w->overthespotWin.roman_width,
      w->overthespotWin.font_height,
      w->overthespotWin.modeshell_popup->core.border_width ) ;

  if( w->overthespotWin.change_minibuffer_font ){
    /* 礭롣*/
    XtResizeWidget
      ( w->overthespotWin.minibuffer_popup,
	w->overthespotWin.minibuf_width *
	w->overthespotWin.roman_width,
	w->overthespotWin.font_height,
	w->overthespotWin.minibuffer_popup->core.border_width ) ;
    /* ߥ˥ХåեΥɥޥ͡㡼°롣*/
    sizehints.base_width  = 
      w->overthespotWin.minibuffer_popup->core.border_width * 2 ;
    sizehints.base_height = 
      w->overthespotWin.minibuffer_popup->core.border_width * 2 ;
    sizehints.width_inc  = w->overthespotWin.roman_width ;
    sizehints.height_inc = w->overthespotWin.font_height ;
    sizehints.min_width  = w->overthespotWin.roman_width * 2 +
      sizehints.base_width ;
    sizehints.min_height = w->overthespotWin.font_height +
      sizehints.base_height ;
    sizehints.max_width  = w->overthespotWin.roman_width * 160 + 
      sizehints.base_width ;
    sizehints.max_height = w->overthespotWin.font_height +
      sizehints.base_height ;
    sizehints.flags = ( PBaseSize | PMinSize | PResizeInc | PMaxSize ) ;
    XChangeWMNormalHints
      ( XtDisplay( gw ),
	XtWindow( w->overthespotWin.minibuffer_popup ),
	&sizehints ) ;
  } else {
    /* 礭롣*/
    XtResizeWidget
      ( w->overthespotWin.minibuffer_popup,
	w->overthespotWin.minibuf_width *
	w->overthespotWin.mroman_width,
	w->overthespotWin.mfont_height,
	w->overthespotWin.minibuffer_popup->core.border_width ) ;
  }
  return ;
}

static void OTSW_ReconfigureColor( Widget gw, unsigned long valuemask )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  Arg args[ 3 ] ;
  Cardinal num_args = 0 ;

  if( valuemask & CAColor ){
    XtSetArg
      ( args[ num_args ], XtNbackground, w->overthespotWin.obackground ) ;
    num_args ++ ;
  }
  if( valuemask & CAColormap ){
    XtSetArg
      ( args[ num_args ], XtNcolormap, w->overthespotWin.colormap ) ;
    num_args ++ ;
  }
  if( num_args > 0 ){
    /*   ֤󡣤ʤꤷʤȤʤΤġ *
     * ե꡼֥եȤ衢եȡ                     */
    XtSetValues( w->overthespotWin.modeshell_popup, args, num_args ) ;
    if( w->overthespotWin.change_minibuffer_font ){
      XtSetValues( w->overthespotWin.minibuffer_popup, args, num_args ) ;
      XtSetValues( w->overthespotWin.minibuffer_canvas, args, num_args ) ;
    }
    XtSetValues( w->overthespotWin.topline_popup, args, num_args ) ;
    XtSetValues( w->overthespotWin.midline_popup, args, num_args ) ;
    XtSetValues( w->overthespotWin.botline_popup, args, num_args ) ;
    XtSetValues( w->overthespotWin.modeshell_canvas, args, num_args ) ;
    XtSetValues( w->overthespotWin.topline_canvas, args, num_args ) ;
    XtSetValues( w->overthespotWin.midline_canvas, args, num_args ) ;
    XtSetValues( w->overthespotWin.botline_canvas, args, num_args ) ;
  }
  if( !( valuemask & CAColor ) )
    return ;

  /* ޤȤäƤ GC Ʋ롣 minibuffer οä *
   * ѹ٤ʡ  */
  OTSW_ReleaseAllGCs( gw ) ;
  OTSW_CreateAllGCs( gw ) ;
  return ;
}

/*
 * åȴ֤ΥåȤؿ
 */
static Boolean OTSW_SetValues
( Widget current, Widget request, Widget new,
  ArgList args, Cardinal *num_args )
{
  OverthespotWinWidget curw = ( OverthespotWinWidget )current ;
  OverthespotWinWidget reqw = ( OverthespotWinWidget )request ;
  OverthespotWinWidget neww = ( OverthespotWinWidget )new ;
  struct ConvAttrsMesg *camsg ;

  /* 񤬽Ƥ顢ľ¹Ԥ롣*/
  if( curw->overthespotWin.jisyo_dirty != reqw->overthespotWin.jisyo_dirty ){
    neww->overthespotWin.jisyo_dirty = reqw->overthespotWin.jisyo_dirty ;
    /* 뤬Ƥʤм¹Ԥʤȡ*/
    if( neww->overthespotWin.modeshell_probe ){
      XtVaSetValues
	( neww->overthespotWin.modeshell_canvas,
	  XtNredrawRequest, True, NULL ) ;
    }
    return False ;
  }
  /* Ѵ°ν׵᤬褿ȸդꡣ*/
  if( reqw->overthespotWin.camsg != NULL ){
    int ret ;
    /* Ѵ°ˤǸ­Τɤ *
     * äƤߤäʤ͡                                     *
     * ա֤Ǥ͡äƤߤޤ礦Ĥ󡪡             */
    camsg = reqw->overthespotWin.camsg ;
    reqw->overthespotWin.camsg = neww->overthespotWin.camsg = 
      curw->overthespotWin.camsg = NULL ;

    ret = OTSW_ChangeAttributes
      ( new, camsg->mask, &camsg->value ) ;

    if( XtIsRealized( new ) )
      OTSW_ReconfigureWindows( new ) ;
    return ret ;
  }
  /* ߥ˥ХåեõȤ˾νԤ*/
  if( reqw->overthespotWin.clearMinibuffer ){
    if( curw->overthespotWin.minibuffer_probe ){
      XtPopdown( curw->overthespotWin.minibuffer_popup ) ;
      curw->overthespotWin.minibuffer_probe =
	neww->overthespotWin.minibuffer_probe = False ;
    }
    reqw->overthespotWin.clearMinibuffer =
      neww->overthespotWin.clearMinibuffer = 
      curw->overthespotWin.clearMinibuffer = False ;
    return FALSE ;
  }
#if 0
  if( reqw->overthespotWin.client_window != None ){
    fprintf( stderr, "Overthespot: Set client window.\n" ) ;
    return FALSE ;
  }
#endif
  if( reqw->overthespotWin.conversion_set_focus ){
    reqw->overthespotWin.conversion_set_focus = 
      neww->overthespotWin.conversion_set_focus = 
      curw->overthespotWin.conversion_set_focus = False ;
    /* modeshell 褵롣*/
    if( !neww->overthespotWin.modeshell_probe ){
      XtPopup( neww->overthespotWin.modeshell_popup, XtGrabNone ) ;
      neww->overthespotWin.modeshell_probe = True ;
    }
    /* mini-buffer ɽ롣⤷ɬפʤСĤǤ뤬*/
    OTSW_ConfigureMinibuffer( new ) ;
    return FALSE ;
  }
  if( reqw->overthespotWin.conversion_unset_focus ){
    reqw->overthespotWin.conversion_unset_focus = 
      neww->overthespotWin.conversion_unset_focus = 
      curw->overthespotWin.conversion_unset_focus = False ;
    if( neww->overthespotWin.modeshell_probe ){
      XtPopdown( neww->overthespotWin.modeshell_popup ) ;
      neww->overthespotWin.modeshell_probe = False ;
    }
    return FALSE ;
  }
  return( FALSE ) ;
}

static int OTSW_ReconfigureFocusPosition
( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int x, y, ofx, ofy ;
  Window dummy ;
  int ret = False ;
  /* 롼ȥɥǤΥեɥа֤򻻽 *
   * 롣*/
  XTranslateCoordinates
    ( XtDisplay( gw ), w->overthespotWin.client_window,
      RootWindowOfScreen( XtScreen( gw ) ),
      0, 0, &x, &y, &dummy ) ;
  if( w->overthespotWin.focus_window != w->overthespotWin.client_window ){
    XTranslateCoordinates
      ( XtDisplay( gw ), w->overthespotWin.focus_window,
	w->overthespotWin.client_window, 
	0, 0, &ofx, &ofy, &dummy ) ;
  } else {
    ofx = ofy = 0 ;
  }
  if( w->overthespotWin.client_x != x || w->overthespotWin.client_y != y ){
    w->overthespotWin.client_x = x ;
    w->overthespotWin.client_y = y ;
    ret = True ;
  }
  if( w->overthespotWin.focus_offset_x != ofx ||
      w->overthespotWin.focus_offset_y != ofy ){
    w->overthespotWin.focus_offset_x = ofx ;
    w->overthespotWin.focus_offset_y = ofy ;
    ret = True ;
  }
  return ret ;
}

/*
 * Ѵ饤ȤΡѴ°ν׵פ򤵤Фؿ
 */
static int OTSW_ChangeAttributes
( Widget gw, unsigned long valuemask, struct ConvAttrs *values )
{
  OverthespotWinWidget w   = ( OverthespotWinWidget )gw ;
  int ret = False, rec_col ;

  if( valuemask & CAStatusArea ){
    /* ɸνä⤷ɽΰ褬Ƥ硣*/
    if( values->status_area.x != w->overthespotWin.status_area.x ||
        values->status_area.y != w->overthespotWin.status_area.y ||
        values->status_area.width  != w->overthespotWin.status_area.width || 
        values->status_area.height != w->overthespotWin.status_area.height ){
      w->overthespotWin.status_area.x = values->status_area.x ;
      w->overthespotWin.status_area.y = values->status_area.y ;
      w->overthespotWin.status_area.width  = values->status_area.width ;
      w->overthespotWin.status_area.height = values->status_area.height ;
      w->overthespotWin.attribute_mask |= CAStatusArea ;
      ret = True ;
    }
  }
  rec_col = False ;
  /* 顼ޥåפѹ׵᤬Ƥ顢롣*/
  if( ( valuemask & CAColormap ) &&
     w->overthespotWin.colormap != values->colormap ){
    w->overthespotWin.colormap = values->colormap ;
#if defined(DEBUG)
    fprintf
      ( stderr, "(OTSW)Change colormap.\n" ) ;
#endif
    rec_col = True ;
  }
  /* ե饦ɥ顼ڤӥХå饦ɥ顼ѹ׵᤬ *
   * 顢ġ*/
  if( ( valuemask & CAColor ) &&
     ( w->overthespotWin.obackground != values->background ||
       w->overthespotWin.oforeground != values->foreground ) ){
#if defined(DEBUG)
    fprintf
      ( stderr, "(OTSW)Change foreground and background colors.\n" ) ;
#endif
    w->overthespotWin.obackground = values->background ;
    w->overthespotWin.oforeground = values->foreground ;
    rec_col = True ;
  }
  if( XtIsRealized( gw ) && rec_col ){
    OTSW_ReconfigureColor( gw, valuemask ) ;
  }
  /* Ѵ饤ȤɽΰνΤ褿ν*/
  if( valuemask & CAClientArea ){
    /* ɸνä⤷ɽΰ褬Ƥ硣*/
    if( values->client_area.x != w->overthespotWin.client_area.x ||
        values->client_area.y != w->overthespotWin.client_area.y ||
        values->client_area.width != w->overthespotWin.client_area.width || 
        values->client_area.height != w->overthespotWin.client_area.height ){
      /* Ѵ饤ȤɽΰѹФƤ*/
      w->overthespotWin.client_area    = values->client_area ;
#ifdef DEBUG
      fprintf
	( stderr, "(OTSW)Client pos:(%d,%d), size:%d x %d.\n",
	  w->overthespotWin.client_area.x,
	  w->overthespotWin.client_area.y,
	  w->overthespotWin.client_area.width,
	  w->overthespotWin.client_area.height ) ;
#endif
      w->overthespotWin.attribute_mask |= CAClientArea ;
      ret = True ;
    }
  }
  if( valuemask & CAFocusWindow ){
#ifdef DEBUG
    fprintf
      ( stderr, "(OTSW)Focus window (%ld)\n", values->focus_window ) ;
#endif
    /* եäƼºߤΡ Ȥ¤Ϥ⤦ä *
     * ɬפȻפ*/
    if( values->focus_window != None ){
      /* ꤷȦΥեʪˤʤäƤ顢̣Ǥ */
      /* ʤ ȤäǤ*/
      if( w->overthespotWin.focus_window != None ){
	XtDestroyWidget( gw ) ;
	return False ;
      }
      w->overthespotWin.focus_window = values->focus_window ;

      /* Ѵ饤ȤΰФ̵ä鼫Ѱդ *
       * 줬ɤʤݾڤϤʤΤɡѴ *
       * Ȥʤ̵ */
      if( !(( w->overthespotWin.attribute_mask ) & CAClientArea )){
	XGetRectangleOfWindow
	  ( XtDisplay( gw ), w->overthespotWin.focus_window,
	    &w->overthespotWin.client_area ) ;
      }
      ret = True ;
    }
  }
  /* ɸȤ׵᤬ƤνԤ*/
  if( valuemask & CASpotLocation ){
    if( w->overthespotWin.focus_window != None ){
      int prev_focus ;
#if defined(DEBUG)
      fprintf( stderr, "CASPOT nanode FOCUSWINDOW miru (%ld).\n",
	      w->overthespotWin.focus_window ) ;
#endif
      prev_focus = w->overthespotWin.is_focus ;
      w->overthespotWin.is_focus = focusCheck( gw ) ;
      if( prev_focus != w->overthespotWin.is_focus ){
	ret = True ;
      }
      if( OTSW_ReconfigureFocusPosition( gw ) )
	ret = True ;
    }
    if( values->spot_x != w->overthespotWin.spot_x ||
	values->spot_y != w->overthespotWin.spot_y ){
#ifdef DEBUG
      fprintf( stderr, "Get Basho Shitei meirei.\n" ) ;
#endif
      /* ֤뤵*/
      w->overthespotWin.spot_x = values->spot_x ;
      w->overthespotWin.spot_y = values->spot_y ;
      ret = True ;
    }
  }
#if 0
 else if( valuemask & CAFocusWindow &&
	    w->overthespotWin.focus_window != None ){
    /* եƤɥưƤ⤳ƤФΤ * 
     * եɤľȤˤ롣kinput2 Ȥɤ *
     * Τʬʤɡ꤬ФƤͤ뤵*/
    XTranslateCoordinates
      ( XtDisplay( gw ), w->overthespotWin.focus_window,
	RootWindowOfScreen( XtScreen( gw ) ),
	0, 0, &x, &y, &child ) ;
    w->overthespotWin.focus_offset_x = x ;
    w->overthespotWin.focus_offset_x = y ;
    ret = True ;
    /*  Ŭ˸߻ȤäƤեȡΤޤޤǤäƤ롣*/
  }
#endif
  /* եȾνäν*/
  if( valuemask & CAFonts ){
#ifdef DEBUG
    fprintf( stderr, "FONT WO KIRIKAERU.\n" ) ;
#endif
    /* եȤڤؤ롣*/
    if( values->fontset[ 0 ] != NULL ){
      w->overthespotWin.fs_roman = values->fontset[ 0 ] ;
    } else {
      /* ꤨʤեȤλ꤬褿ˤϡǥեȤ᤹*/
      w->overthespotWin.fs_roman = w->overthespotWin.fs_default_roman ;
    }
    if( values->fontset[ 1 ] != NULL ){
      w->overthespotWin.fs_kanji = values->fontset[ 1 ] ;
    } else {
      /* ꤨʤեȤλ꤬褿ˤϡǥեȤ᤹*/
      w->overthespotWin.fs_kanji = w->overthespotWin.fs_default_kanji ;
    }
    /* եȾ롣*/
    getFontInfo
      ( w->overthespotWin.fs_kanji, w->overthespotWin.fs_roman,
	&w->overthespotWin.kanji_width, &w->overthespotWin.roman_width,
	&w->overthespotWin.font_height, &w->overthespotWin.font_ascent ) ;
    /* ޤåȤβƤˤϴ GC ȤѰդ */
    /* ƤޤäƤΤǡʤФʤʤ*/ 
    if( XtIsRealized( gw ) ){
      /* ߥ˥Хåեȥ⡼ɥ礭롣*/
      OTSW_ResizeMinibufferAndModeshell( gw ) ;
      /* ޤȤäƤ GC Ʋ롣*/
      OTSW_ReleaseAllGCs( gw ) ;
      /* ľ*/
      OTSW_CreateAllGCs( gw ) ;
    }
    ret = True ;
  }
#undef XQueryRaiseWindow

  return ret ;
}

enum {
  SIW_DESTROYED = -1, SIW_PROCESSING = 0,
} ;

/*
 * 줿Ȥ٥ȤνԤؿ
 *----
 * Υ줿ȤԲǽʾ礬Τǡ줿Ȥ
 * Ǿ꤯ư褦˺ʤФʤʤ
 */
static void KeyDownEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  char inpstr[ STRBUFSIZE ] ;
  KeySym key ;
  struct SKKInputNode *node, *topbuffer ;

  /* 줿ΤǤʤȴ롣*/
  if( xevent->type != KeyPress )
    return ;

  /* ٥Ȥ򵭲Ƥ*/
  w->overthespotWin.xevent = *xevent ;

  /* ߥ˥Хåե¸ߤƤΤʤСΰֺǸΤΤ򤹤롣*/
  topbuffer = w->overthespotWin.topbuffer ;
  node      = w->overthespotWin.lastbuffer ;
  if( !node->cur_exist )
    node = topbuffer ;

  /* ʸХåեν*/
  inpstr[ 0 ] = '\0' ;
  /* ʸ*/
  XLookupString( &( xevent->xkey ), inpstr, STRBUFSIZE, &key, NULL ) ;
  
  /* եȥڡξνԤ*/
  if( ( xevent->xkey.state & ShiftMask ) && key == XK_space ){
    XtDestroyWidget( gw ) ;
    return ;
  }
  
  /* Ϥ줿ʸϰ̣ʸʤΤǤ */
  if( inpstr[ 0 ] != '\0' ){
    w->overthespotWin.keybuf[ w->overthespotWin.keybuf_use ++ ] =
      inpstr[ 0 ] ;
    /* ߥ˥Хåե¸ߤѤƤν*/
    switch( do_function_by_reading_character( gw, node ) ){
    case SIW_DESTROYED :
      XtDestroyWidget( gw ) ;
      return ;
    case SIW_PROCESSING :
      break ;
    default :
      return ;
    }
  } else {
    /* ȥ륹ڡξνԤ string  */
    /* ʤȦʤΤǡ*/
    if( ( xevent->xkey.state & ControlMask ) && key == XK_space ){
      w->overthespotWin.keybuf[ w->overthespotWin.keybuf_use ++ ] = 0x00 ;
      if( do_function_by_reading_character( gw, node ) )
	return ;
      OTSW_RedrawScreen( gw ) ;
      return ;
    }
    /* skkinput ǤǧǤʤ̣뤫⤷ʤ֤*/
    if( topbuffer->textbuffer[ 0 ] == '\0' &&
        topbuffer->minibuffer == NULL ){
      j_sendback_keypress( gw, &( w->overthespotWin.xevent ) ) ;
    }
  }
  /*skkinput_DrawScreen( gw ) ;*/
  /* ̤ľĺʬǤΤʤ*/
  OTSW_RedrawScreen( gw ) ;
/*  XFlush( XtDisplay( gw ) ) ;*/
  return ;
}

/*
 * ߥޥեƤ뤫ɤå뤿δؿ
 *-----
 * ΤƤХȤơեΰư˾꤯ȡ
 * Pointer ¸ߤƤ뤳ȤƨƤޤȤġʤʬ
 * 뤫
 */
static int focusCheck( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  Window mouse_focus, rootwin, childwin ;
  int ret_focus_state ;
  int rootx, rooty, winx, winy ;
  unsigned int winmask ;
  
  /* ޤޥΥեåäƤͤդ롣*/
  XGetInputFocus
    ( XtDisplay( gw ), &mouse_focus, &ret_focus_state ) ;

#if defined(DEBUG)
  fprintf( stderr, "Mouse Focus Window:(%ld)\n", mouse_focus ) ;
#endif
  if( mouse_focus == None || mouse_focus == PointerRoot )
    return False ;

  if( w->overthespotWin.input_focus_window != None ){
    if( mouse_focus == w->overthespotWin.input_focus_window )
      return True ;
  } 
  if( XQueryPointer
     ( XtDisplay( gw ), mouse_focus,
       &rootwin, &childwin, &rootx, &rooty, &winx, &winy, 
       &winmask ) ){
#if defined(DEBUG)
    fprintf
      ( stderr, "Focus Root Window:(%ld), Focus Child Window:(%ld)\n",
        rootwin, childwin ) ;
#endif
    /* Ѵ饤ȤޥΥեäƤ롢⤷ */
    /* ϤλҶޥΥեäƤΤʤ True */
    if( w->overthespotWin.focus_window == childwin ){
      if( w->overthespotWin.input_focus_window == None ){
	w->overthespotWin.input_focus_window = mouse_focus ;
      }
      return True ;
    }
    if( w->overthespotWin.focus_window == mouse_focus )
      return True ;
    /* ߥ˥ХåեɥƤ⤢衣*/
    if( w->overthespotWin.minibuffer_probe ){
      if( XtWindow( w->overthespotWin.minibuffer_popup ) == childwin )
	return True ;
      if( XtWindow( w->overthespotWin.minibuffer_popup ) == mouse_focus )
	return True ;
    }
  }
  return False ;
}

/*
 * ե蘆줿νԤؿ
 */
static void FocusEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int prev_focus ;

  /* եɥޤäƤʤΤˡʥեIN ʤ */
  /* äƤ礸ʤ*/
  if( w->overthespotWin.focus_window == None )
    return ;
  /* ޤϥե򼺤äΤȹͤ롣*/
  prev_focus = w->overthespotWin.is_focus ;
  w->overthespotWin.is_focus = focusCheck( gw ) ;
  /* ̤νĥηѲΤǡ*/
  if( prev_focus != w->overthespotWin.is_focus )
    OTSW_RedrawScreen( gw ) ;
  return ;
}

/*
 * ʲΥեϡOverthespotWinWidget ΥΰǤ롣
 */

/*
 * ơ֥򻲾Ȥ̤Υե󥯥ֹ˽äƽ¹Ԥ
 * ٤¾δؿƽФԤؿ
 */
static int do_skkinputFunction_withtext
( Widget gw, struct SKKInputNode *node,	int func_no )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int chara = w->overthespotWin.keybuf[ 0 ] ;

  /* ȡΥåХåե˽񤫤Ƥ *
   * Ͼä뤳ȤˤʤäƤ롩 */
  node->mtextbuffer[ 0 ] = '\0' ;

  /* 䴰ײλ뤫ɤ*/
  if( node->j_completion_mode ){
    if( func_no != FUNCNO_J_INSERT_PERIOD &&
	func_no != FUNCNO_J_INSERT_COMMA  &&
	func_no != FUNCNO_J_ABBREV_PERIOD &&
	func_no != FUNCNO_J_ABBREV_COMMA  ){
      j_completion_close( node ) ;
    }
  }
  /* Ϳ줿̿ФԤ*/
  switch( func_no ){
  case FUNCNO_J_PREFIX_CHAR :
    j_delete_kanaprefix( node ) ;
    /* ֢ΥХåեʤСC- Ȥɽ롣*/
    if( node == ( w->overthespotWin ).topbuffer ){
      strcpy( node->mtextbuffer,
	     charaToString( w->overthespotWin.keybuf[ 0 ] ) ) ;
      strcat( node->mtextbuffer, "-" ) ;
    }
    return SIW_PROCESSING ;
  case FUNCNO_SELF_INSERT_CHARACTER :
    j_insert_chara( node, node->cur_pos, chara ) ;
    break ;
  case FUNCNO_J_SELF_INSERT :
    j_self_insert
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_SELF_ZENKAKU_INSERT :
    j_insert_zenkaku( node, node->cur_pos, chara ) ;
    break ;
  case FUNCNO_J_DISPLAY_CODE_FOR_CHAR_AT_POINT :
    j_display_code_for_char_at_point( node ) ;
    break ;
  case FUNCNO_J_SET_HENKAN_POINT :
    /* Ѵϰ֤ꤵ롣*/
    j_set_henkan_point
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_SET_HENKAN_POINT_SUBR :
    j_set_henkan_point_subr
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
    /* 첻ϡ*/
  case FUNCNO_J_INSERT_A :
    j_insert_a
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_E :
    j_insert_e
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_I :
    j_insert_i
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_O :
    j_insert_o
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_U :
    j_insert_u
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
    /* ̾Ϥγϥ( ҲǤ )*/
  case FUNCNO_J_KANA_INPUT :
    j_kana_input
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_START_HENKAN :
    j_start_henkan
      ( gw, &w->overthespotWin.topbuffer,
	&w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INSERT_COMMA :
    j_insert_comma
      ( gw, &w->overthespotWin.topbuffer, 
	&w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INSERT_PERIOD :
    j_insert_period
      ( gw, &w->overthespotWin.topbuffer,
	&w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_PURGE_FROM_JISYO :
    j_purge_from_jisyo( gw, &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INPUT_BY_CODE_OR_MENU :
    j_input_by_code_or_menu_start
      ( gw, &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_MODE_OFF :
    j_mode_off
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_MODE_OFF_AND_SELF_INSERT :
    j_mode_off_and_self_insert
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_TOGGLE_KANA :
    /* ̾⡼ɤҲ̾⡼ɤڤؤ*/
    j_toggle_kana
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, False, chara ) ;
    break ;
  case FUNCNO_J_PREVIOUS_CANDIDATE :
    j_previous_candidate
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_KAKUTEI :
    if( j_kakutei_kana_mode
	( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	  node, chara ) )
      break ;
    j_kakutei( gw, node ) ;
    break ;
  case FUNCNO_J_ABBREV_INPUT :
    j_abbrev_input
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_ABBREV_PERIOD :
    j_abbrev_period( gw, node ) ;
    break ;
  case FUNCNO_J_ABBREV_COMMA :
    j_abbrev_comma( gw, node ) ;
    break ;
  case FUNCNO_J_ZENKAKU_EIJI :
    /* (Ѥܸ)ѻϥ⡼ɤˤ롣*/
    j_zenkaku_eiji
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_ZENKAKU_HENKAN :
    j_zenkaku_henkan( gw, node ) ;
    break ;
  case FUNCNO_J_TODAY :
    /* դνϡ*/
    j_today( node ) ;
    break ;
  case FUNCNO_J_KANAINPUT_MODE_ON :
    j_kanainput_mode_on
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_SET_MARK_COMMAND :
    j_set_mark_command( node ) ;
    break ;
  case FUNCNO_J_NEWLINE :
    j_newline
      ( gw, &( w->overthespotWin.xevent ), True,
	&w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, w->overthespotWin.egg_like_newline,
	False,
	NULL, NULL, NULL, NULL ) ;
    break ;
  case FUNCNO_J_FORWARD_CHAR :
    j_forward_char( node ) ;
    break ;
  case FUNCNO_J_BACKWARD_CHAR :
    j_backward_char( node ) ;
    break ;
  case FUNCNO_J_DELETE_CHAR :
    j_delete_char( node ) ;
    break ;
  case FUNCNO_J_DELETE_BACKWARD_CHAR :
    j_backward_delete_char
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node,
        &w->overthespotWin.xevent, True ) ;
    break ;
  case FUNCNO_J_TRY_COMPLETION :
    j_try_completion( gw, node ) ;
    break ;
  case FUNCNO_J_END_OF_LINE :
    j_end_of_line( node ) ;
    break ;
  case FUNCNO_J_BEGINNING_OF_LINE :
    j_beginning_of_line( node ) ;
    break ;
  case FUNCNO_J_KILL_LINE :
    j_kill_line( node, w->overthespotWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_KILL_REGION :
    j_kill_region( node, w->overthespotWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_KILL_RING_SAVE :
    j_kill_ring_save( node, w->overthespotWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_EXCHANGE_POINT_AND_MARK :
    j_exchange_point_and_mark( node ) ;
    break ;
  case FUNCNO_J_YANK :
    j_yank( node, w->overthespotWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_PREVIOUS_LINE :
    break ;
  case FUNCNO_J_NEXT_LINE :
    break ;
  case FUNCNO_J_TRANSPOSE_CHARS :
    j_transpose_chars( node ) ;
    break ;
  case FUNCNO_J_REDRAW :
    /* öե褹롣ʬǤϤʤơ*/
    OTSW_RedrawScreen( gw ) ;
    break ;
  case FUNCNO_J_KEYBOARD_QUIT :
    j_keyboard_quit
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_SAVE_SKKINPUT_LOCAL_JISYO :
    j_delete_kanaprefix( node ) ;
    /* 񤬹Ƥ뤫ɤȽꤹ롣*/
    if( w->overthespotWin.jisyo_dirty ){
      strcpy( node->mtextbuffer, "Saving Jisyo..." ) ;
      /*skkinput_DrawScreen( gw ) ;*/
      OTSW_RedrawScreen( gw ) ;
      /* ι򤹤륳ХåƽФ*/
      skkinput_updateLocalJisyo() ;
      XSync( XtDisplay( gw ), True ) ;
      
      strcat( node->mtextbuffer, "done" ) ;
    } else {
      err_msg( node, "(No changes need to be saved)" ) ;
    }
    break ;
  case FUNCNO_CLOSE_SKKINPUT :
    return SIW_DESTROYED ;
  case FUNCNO_VC_TOGGLE_CHATMODE :
    j_delete_kanaprefix( node ) ;
    w->overthespotWin.chat_adapter =
      ( w->overthespotWin.chat_adapter )? False : True ;
    break ;
  case FUNCNO_VC_TOGGLE_EGGNL :
    j_delete_kanaprefix( node ) ;
    w->overthespotWin.egg_like_newline =
      ( w->overthespotWin.egg_like_newline )? False : True ;
    break ;
  case FUNCNO_NOT_MODIFIED :
    j_delete_kanaprefix( node ) ;
    /* ѹե饰õ롣*/
    j_clear_modified_flag() ;
    strcpy( node->mtextbuffer, "Modification-flag cleared" ) ;
    break ;
  case FUNCNO_J_FORWARD_WORD :
    j_forward_word( node ) ;
    break ;
  case FUNCNO_J_BACKWARD_WORD :
    j_backward_word( node ) ;
    break ;
  case FUNCNO_J_UPCASE_WORD :
    j_caseword( node, 'u' ) ;
    break ;
  case FUNCNO_J_DOWNCASE_WORD :
    j_caseword( node, 'd' ) ;
    break ;
  case FUNCNO_J_CAPITALIZE_WORD :
    j_caseword( node, 'c' ) ;
    break ;
  case FUNCNO_INVALID_CHAR :
  default :
#if 0
    XBell( XtDisplay( gw ), 5 ) ;
#endif
    break ;
  }
  w->overthespotWin.keybuf_use  = 0 ;
  w->overthespotWin.keybuf[ 0 ] = '\0' ;

  return SIW_PROCESSING ;
}

/*
 * ơ֥򻲾Ȥ̤Υե󥯥ֹ˽äƽ¹Ԥ
 * ٤¾δؿƽФԤؿ
 */
static int do_skkinputFunction_withouttext
( Widget gw, struct SKKInputNode *node,	int func_no )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int chara = w->overthespotWin.keybuf[ 0 ] ;

  /* ȡΥåХåե˽񤫤Ƥ *
   * Ͼä뤳ȤˤʤäƤ롩 */
  node->mtextbuffer[ 0 ] = '\0' ;

  /* 褿ɬ completion Ͻλ뤳Ȥˤʤ롣*/
  if( node->j_completion_mode ){
    j_completion_close( node ) ;
  }
  /* Ϳ줿̿ФԤʸʳ̵뤹롣*/
  switch( func_no ){
  case FUNCNO_SELF_INSERT_CHARACTER :
    j_insert_chara( node, node->cur_pos, chara ) ;
    break ;
  case FUNCNO_J_SELF_INSERT :
    j_self_insert
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_SELF_ZENKAKU_INSERT :
    j_insert_zenkaku( node, node->cur_pos, chara ) ;
    break ;
  case FUNCNO_J_SET_HENKAN_POINT :
    /* Ѵϰ֤ꤵ롣*/
    j_set_henkan_point
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_SET_HENKAN_POINT_SUBR :
    j_set_henkan_point_subr
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
    /* 첻ϡ*/
  case FUNCNO_J_INSERT_A :
    j_insert_a
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_E :
    j_insert_e
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_I :
    j_insert_i
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_O :
    j_insert_o
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_U :
    j_insert_u
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_PURGE_FROM_JISYO :
    break ;
    /* ̾Ϥγϥ( ҲǤ )*/
  case FUNCNO_J_KANA_INPUT :
    j_kana_input
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_START_HENKAN :
    j_start_henkan
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INSERT_COMMA :
    j_insert_comma
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INSERT_PERIOD :
    j_insert_period
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INPUT_BY_CODE_OR_MENU :
    j_input_by_code_or_menu_start
      ( gw, &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_MODE_OFF :
    j_mode_off
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node, chara ) ;
    /* Ϻƹͤ;Ϥꡣ˽롣*/
    break ;
  case FUNCNO_J_MODE_OFF_AND_SELF_INSERT :
    j_mode_off_and_self_insert
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_TOGGLE_KANA :
    /* ̾⡼ɤҲ̾⡼ɤڤؤ*/
    j_toggle_kana
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, False, chara ) ;
    break ;
  case FUNCNO_J_PREVIOUS_CANDIDATE :
    j_previous_candidate
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_KAKUTEI :
    if( j_kakutei_kana_mode
	( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	  node, chara ) )
      break ;
    j_kakutei( gw, node ) ;
    break ;
  case FUNCNO_J_ABBREV_INPUT :
    j_abbrev_input
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_ABBREV_PERIOD :
    j_abbrev_period( gw, node ) ;
    break ;
  case FUNCNO_J_ABBREV_COMMA :
    j_abbrev_comma( gw, node ) ;
    break ;
  case FUNCNO_J_ZENKAKU_EIJI :
    /* (Ѥܸ)ѻϥ⡼ɤˤ롣*/
    j_zenkaku_eiji
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_TODAY :
    /* դνϡ*/
    j_today( node ) ;
    break ;
  case FUNCNO_J_KANAINPUT_MODE_ON :
    j_kanainput_mode_on
      ( gw, &w->overthespotWin.topbuffer,
        &w->overthespotWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_NEWLINE :
    j_newline
      ( gw, &( w->overthespotWin.xevent ), True,
	&w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, w->overthespotWin.egg_like_newline, False,
	NULL, NULL, NULL, NULL ) ;
    break ;
  case FUNCNO_CLOSE_SKKINPUT :
    return SIW_DESTROYED ;
  case FUNCNO_VC_TOGGLE_EGGNL :
    j_delete_kanaprefix( node ) ;
    w->overthespotWin.egg_like_newline =
      ( w->overthespotWin.egg_like_newline )? False : True ;
    break ;
  case FUNCNO_SAVE_SKKINPUT_LOCAL_JISYO :
    j_delete_kanaprefix( node ) ;
    /* 񤬹Ƥ뤫ɤȽꤹ롣*/
    if( w->overthespotWin.jisyo_dirty ){
      strcpy( node->mtextbuffer, "Saving Jisyo..." ) ;
      /*skkinput_DrawScreen( gw ) ;*/
      OTSW_RedrawScreen( gw ) ;
      /* ι򤹤륳ХåƽФ*/
      skkinput_updateLocalJisyo() ;
      /* ä٥ȤϼʤΤä롣*/
      XSync( XtDisplay( gw ), True ) ;
      
      strcat( node->mtextbuffer, "done" ) ;
    } else {
      err_msg( node, "(No changes need to be saved)" ) ;
    }
    break ;
    /* ϥ֤Ȥˤʤ롣*/
  case FUNCNO_INVALID_CHAR :
  default :
    j_sendback_keypress( gw, &( w->overthespotWin.xevent ) ) ;
    break ;
  }
  w->overthespotWin.keybuf_use  = 0 ;
  w->overthespotWin.keybuf[ 0 ] = '\0' ;

  return SIW_PROCESSING ;
}

/*
 * ĤξνԤؿ
 */
static int do_function_by_onekey
( Widget gw, struct SKKInputNode *node, int textflag )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int func_no ;

  /* Ѵθ⡼ɤʤΤǤ */
  if( node->j_henkan_show_candidate_mode ){
    /* ɤʥäϤͤȤäɡĤޤΥ⡼*
     * ɤФäƤ뤫פʡ */
    do_Function_jhenkanShowCandidateMode
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, w->overthespotWin.keybuf ) ;
    /* ʸΥХåեõ롣*/
    w->overthespotWin.keybuf_use  = 0 ;
    w->overthespotWin.keybuf[ 0 ] = '\0' ;
    return SIW_PROCESSING ;

  } else if( node->j_input_by_code_or_menu_mode ){
    if( node->j_input_by_code_or_menu_mode == True ){
      do_function_j_input_by_code_or_menu_jump
	( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	  node, w->overthespotWin.keybuf ) ;
      /* ʸΥХåեõ롣*/
      w->overthespotWin.keybuf_use  = 0 ;
      w->overthespotWin.keybuf[ 0 ] = '\0' ;
      return SIW_PROCESSING ; 

    } else {
      do_function_j_input_by_code_or_menu_1_jump
	( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	  node, w->overthespotWin.keybuf ) ;
      /* ʸΥХåեõ롣*/
      w->overthespotWin.keybuf_use  = 0 ;
      w->overthespotWin.keybuf[ 0 ] = '\0' ;
      return SIW_PROCESSING ; 
    }
  }
  /* Ϥ줿ǽʸ򸫤ơĵǽʤǤäФ*/
  func_no = lookup_at_keymap( w->overthespotWin.keybuf, node ) ;

  /* 
   * ¿ʬʳ skk-rom-kana-rule-list åʤȤ
   * ǤʤХȥ륳ɤǤ⤽줬ȤȤȤ
   * Ϥʤʤ顣
   */
  switch( func_no ){
  case FUNCNO_J_INSERT_A :
  case FUNCNO_J_INSERT_E :
  case FUNCNO_J_INSERT_I :
  case FUNCNO_J_INSERT_O :
  case FUNCNO_J_INSERT_U :
  case FUNCNO_J_KANA_INPUT :
  case FUNCNO_J_SET_HENKAN_POINT :
    break ;
  default :
    /* ̾Ϥäꤷˤϡ⤷   *
     * rom-kana-rule-list ˥ҥåȤꤹ뤫⤷ʤ*/
    if( node->j_kana_mode ){
      if( rom_kana_rule_list_check
	  ( node, w->overthespotWin.keybuf[ 0 ] ) ){
	if( node->j_henkan_mode && !node->j_henkan_on &&
	    node->j_okurigana_mode ){
	  j_set_okurigana
	    ( gw, &w->overthespotWin.topbuffer,
	      &w->overthespotWin.lastbuffer,
	      node, node->j_okuri_chara ) ;
	}
	w->overthespotWin.keybuf_use  = 0 ;
	w->overthespotWin.keybuf[ 0 ] = '\0' ;
	return SIW_PROCESSING ;
      }
    }
    /*  oh Υå롣*/
    j_oh_check_with_controlcode
      ( node, ( ( func_no == FUNCNO_J_KEYBOARD_QUIT )? True : False ) ) ;
    break ;
  }
  /* Ϳ줿ǽ¹פޤ*/
  if( textflag ){
    return do_skkinputFunction_withtext( gw, node, func_no ) ;
  } else {
    return do_skkinputFunction_withouttext( gw, node, func_no ) ;
  }
}

/*
 * ĤξνԤؿ
 */
static int do_function_by_twokeys
( Widget gw, struct SKKInputNode *node, int textflag )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;

  /* ȡΥåХåե˽񤫤Ƥ *
   * Ͼä뤳ȤˤʤäƤ롩 */
  node->mtextbuffer[ 0 ] = '\0' ;

  /* Ϥ줿Ϥ⤷ơ */
  if( lookup_at_keymap( w->overthespotWin.keybuf, node ) ==
      FUNCNO_J_SENDBACK_KEY ){
#if 0
    /* Ȥ򤷤Ƥ⡢spot_x, spot_y νäʤΤ *
     * ˤʤˤʤäƤޤover-the-spot mode ǤϽʤ褦 *
     * 롣*/
    j_sendback_keypress( gw, &( w->overthespotWin.xevent ) ) ;
#else
    /* ʾʸФʤΤǡľ*/
    w->overthespotWin.keybuf[ 0 ] = '\0' ;
    w->overthespotWin.keybuf_use  = 0 ;
    return SIW_PROCESSING ;
#endif
  } else {
    /* ơ֥ܤʣξϤʤ *
     * õʤƴŤȤޤ*/
    if( textflag ){
      return do_skkinputFunction_withtext
	( gw, node,
	  lookup_at_twokeymap( w->overthespotWin.keybuf ) ) ;
    } else {
      return do_skkinputFunction_withouttext
	( gw, node,
	  lookup_at_twokeymap( w->overthespotWin.keybuf ) ) ;
    }
      
  }
  /* ʾʸФʤΤǡľ*/
  w->overthespotWin.keybuf[ 0 ] = '\0' ;
  w->overthespotWin.keybuf_use  = 0 ;
  return SIW_PROCESSING ;
}

/*
 * Ϥ줿ʸ󤫤餽б뵡ǽ¹Ԥؿ
 */
static int do_function_by_reading_character
( Widget gw, struct SKKInputNode *node )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  struct SKKInputNode *topbuffer = w->overthespotWin.topbuffer ;
  int textflag ;
  int ret = SIW_PROCESSING ;

#ifdef DEBUG
  fprintf( stderr, "j-prefix : %s\n", node->j_prefix ) ;
#endif

  /* ˥ɥƤ뤫ɤȽǤ롣Ѵưʤ顢 *
   * Ǥư skkinput ôʤФʤʤʳ *
   * ۤȤɤΥ֤ȤˤʤΤǡ*/
  if( topbuffer->textbuffer[ 0 ] != '\0' || 
      topbuffer->minibuffer != NULL ){
    textflag = True ;
  } else {
    textflag = False ;
  }
  /* äϡѴ饤ȤؤۤȤɤΥ뤳Ȥˤʤ롣*/
  switch( w->overthespotWin.keybuf_use ){
    /* ХϤФ*/
  case 1:
    ret = do_function_by_onekey( gw, node, textflag ) ;
    break ;
    /* ʣХȤΥϤФ*/
  case 2:
    ret = do_function_by_twokeys( gw, node, textflag ) ;
    break ;
    /* ꤨʤϤФ*/
  case 0 :
  default :
    w->overthespotWin.keybuf_use  = 0 ;
    w->overthespotWin.keybuf[ 0 ] = '\0' ;
    break ;
  }
  /* Ϥ줿ʸѴ饤Ȥؤ٤ɤȽǤ롣*/
  if( topbuffer->j_henkan_mode || topbuffer->j_kana_mode ||
      topbuffer->j_henkan_on || topbuffer->textbuffer[ 0 ] == '\0' ||
      topbuffer->minibuffer != NULL ||
      topbuffer->mtextbuffer[ 0 ] != '\0' ){
    /* Ѵư桢ޤϡ̾ϤǤ뤫顢ʸ *
     * Ѵ饤Ȥؤޤ櫓ˤϤʤ*/
    return ret ;
  }
  /*  */
  XFlush( XtDisplay( gw ) ) ;
  /* Ѵ饤ȤظϤƤʸ롣*/
  XtCallCallbacks( gw, XtNfixNotify, ( XtPointer )topbuffer->textbuffer ) ;
  /* ƥȥХåե롣*/
  topbuffer->textbuffer[ 0 ] = '\0' ;
  /* ƬؤȰư롣*/
  topbuffer->cur_pos = 0 ;
  topbuffer->j_okurigana_start_point = -1 ;
  topbuffer->j_kana_start_point      = -1 ;
  topbuffer->j_henkan_start_point    = -1 ;
  topbuffer->j_henkan_end_point      = -1 ;
  return ret ;
}

enum {
  MODELINE_NOTHING = 0, MODELINE_HANKAKU, MODELINE_ZENKAKU,
  MODELINE_KATAKANA, MODELINE_KANA,
} ;

/*
 * Ϥ줿 skkinput Υ⡼ɥ饤󤬲ǤΤĴ٤ؿ
 */
static int skkinput_GetModelineNumber( struct SKKInputNode *node )
{
  int modeline ;
  /* ̾ϥ⡼ɤʤΡ ǤϤʤΡ */
  if( !node->j_mode ){
    /* ϥ⡼ɤʤΡ Ⱦϥ⡼ɤʤΡ */
    if( node->j_zenkaku ){
      modeline = MODELINE_ZENKAKU ;
    } else {
      modeline = MODELINE_HANKAKU ;
    }
  } else {
    /* Ҳ̾ϥ⡼ɤʤΡ ʿ̾ϥ⡼ɤʤΡ */
    if( node->j_katakana_mode ){
      modeline = MODELINE_KATAKANA ;
    } else {
      modeline = MODELINE_KANA ;
    }
  }
  return modeline ;
}

/*
 * ⡼ɥ饤褹ؿ
 */
static void OTSW_FullDrawModeLine
( Display *disp, Window win, GC roman_ngc, GC kanji_ngc, 
  struct SKKInputNode *node, int x, int y,
  int roman_width, int kanji_width, int jisyo_dirty )
{
  int modeline = skkinput_GetModelineNumber( node ) ;
  static char *modelineStrings[] = {
    "SKK:", "", "", "",
  } ;
  /* νˤäư㤦⡼ɥ饤*/
  static char *modelineStrings2[] = {
    "--", "*-",
  } ;
  /* ⡼ɥ饤ɽ( SKKΥ⡼ɽ )*/
  skkwin_jputstring
    ( disp, win, roman_ngc, kanji_ngc, 
      x, y, modelineStrings[ modeline - 1 ],
      roman_width, kanji_width ) ;
  /* ֤ι*/
  x += roman_width * 4 ;
  /* ⡼ɥ饤ɽ( ξɽ )*/
  skkwin_jputstring
    ( disp, win, roman_ngc, kanji_ngc,
      x, y, modelineStrings2[ jisyo_dirty ],
      roman_width, kanji_width ) ;
  return ;
}

/*
 * ιԤ褹ؿ
 *------
 * ʣԽ񤫤ʤФʤʤ⤷ʤΤǡ褹٤Կ
 * Ϥɬפ롣
 */
static int OTSW_FullDrawTopBufferMiddleLines
( Display *disp, Window win,
  GC roman_ngc, GC kanji_ngc, GC roman_rgc, GC kanji_rgc,
  unsigned char *text, int text_pos,
  int roman_width, int kanji_width, int font_ascent, int font_height,
  int width, int line_num, int cur_pos, int cur_exist, int is_focus )
{
  int x = 0, y = 0 ;
  int kanji_check, kana_flag, chara = 0, cursor_flag = False ;
  
  kanji_check = kana_flag = False ;

  for( ; *text != '\0' ; text ++, text_pos ++ ){
    /* ɤȽԤ*/
    kanji_check = is_euc_kanji( *text ) ;
    /* ɽʤΤɤȽꤹ롣*/
    if( text_pos == cur_pos && cur_exist ){
      if( ( kanji_check && ( x + kanji_width ) < width ) ||
	 ( !kanji_check && ( x + roman_width ) < width ) ){
	skkinputDrawCursor
	  ( disp, win, roman_ngc, roman_rgc,
	   x, y + font_ascent, font_ascent,
	   ( kanji_check )? kanji_width : roman_width,
	   font_height, is_focus ) ;
	cursor_flag = True ;
      }
    }
    /* ɤΰХܤʤΤǤ礦 */
    if( kanji_check && kana_flag == False ){
      kana_flag = True ;
      chara = ( unsigned char )*text ;
      /* ̤αüޤ褷ΤɤȽǤ롣*/
      if( ( x + kanji_width ) >= width ){
	skkwin_jputchar
	  ( disp, win, roman_ngc, kanji_ngc, 
	    x, y + font_ascent, '\\', False ) ;
	skkwin_jputchar
	  ( disp, win, roman_ngc, kanji_ngc,
	    x + roman_width, y + font_ascent, '\\', False ) ;
	y += font_height ;
	x = 0 ;
	line_num -- ;
	if( line_num <= 0 )
	  break ;
      }
      continue ;
    }
    /* ɤХܤʤΤǤ礦 */
    if( kana_flag ){
      chara = ( chara << 8 ) | ( ( unsigned char )*text ) ;
      if( cursor_flag ){
	if( is_focus ){
	  skkwin_jputchar
	    ( disp, win, roman_rgc, kanji_rgc,
	     x, y + font_ascent, chara, True ) ;
	} else {
	  skkwin_jputchar
	    ( disp, win, roman_ngc, kanji_ngc,
	     x, y + font_ascent, chara, True ) ;
	}
	cursor_flag = False ;
      } else {
	skkwin_jputchar
	  ( disp, win, roman_ngc, kanji_ngc,
	   x, y + font_ascent, chara, False ) ;
      }
      x += kanji_width ;
      kana_flag = False ;
      continue ;
    }
    /* ɤäνǤ*/
    if( ( x + roman_width ) >= width ){
      skkwin_jputchar
	( disp, win, roman_ngc, kanji_ngc,
	  x, y + font_ascent, '\\', False ) ;
      x = 0 ;
      y += font_height ;
      line_num -- ;
      if( line_num <= 0 )
	break ;
    }
    if( cursor_flag ){
      if( is_focus ){
	skkwin_jputchar
	  ( disp, win, roman_rgc, kanji_rgc,
	   x, y + font_ascent, *text, True ) ;
      } else {
	skkwin_jputchar
	  ( disp, win, roman_ngc, kanji_ngc,
	   x, y + font_ascent, *text, True ) ;
      }
      cursor_flag = False ;
    } else {
      skkwin_jputchar
	( disp, win, roman_ngc, kanji_ngc,
	 x, y + font_ascent, *text, False ) ;
    }
    x += roman_width ;
    /* ԥɤξˤϡɽ֤ɬפ롣*/
    if( *text == '\n' ){
      y += font_height ;
      x = 0 ;
      line_num -- ;
      if( line_num <= 0 )
	break ;
    }
  }
  return text_pos ;
}

/*
 * ̵礭ƱʤؿˤʤääȽɬפ
 * 롣
 */
static void OTSW_CalcSizeOfBufferWindows( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int x, y, width, height, kanji_check, kana_flag ;
  struct SKKInputNode *node ;
  unsigned char *text ;
  unsigned char *line_top, *midline_top ;
  int cursor_flag, pos, chara, linetop_pos, line_num, midline_pos ;

#define XQueryPopdown(w,canvas,popupw) \
do { \
       if( w->overthespotWin.canvas ){ \
			    XtPopdown(w->overthespotWin.popupw) ; \
			    w->overthespotWin.canvas = False ; \
			} \
   } while( 0 )
      
  /* ޤɸ򥻥åȤƤ*/
  x = w->overthespotWin.spot_x ;
  y = w->overthespotWin.spot_y ;
  /* ɽΰ礭ȤΤ׻Ƥ*/
  width  = w->overthespotWin.client_area.width +
    w->overthespotWin.client_area.x ;
  height = w->overthespotWin.client_area.height +
    w->overthespotWin.client_area.y ;
  /* Ĵ٤ʤФʤʤʸƬˤ碌Ƥ*/
  node = w->overthespotWin.topbuffer ;
  text = node->textbuffer ;

  if( *text == '\0' ){
    /* ʸ󤬰ĤäƤʤä顢ХåեȤɽ뤿 *
     * ʤƤΤפˤʤΤ顢äƤޤλˡ *
     * ¸ߤƤʤ褦äƤޤܤƤޤΤ *
     * դ뤳 */
    XQueryPopdown( w, topline_probe, topline_popup ) ;
    XQueryPopdown( w, midline_probe, midline_popup ) ;
    XQueryPopdown( w, botline_probe, botline_popup ) ;
    w->overthespotWin.canvas_lines = 0 ;
    return ;
  }

  line_top    = text ;
  linetop_pos = 0 ;
  line_num    = 0 ;
  midline_top = NULL ;

  /* Ĵ٤뤳Ȥˤʤ롣*/
  for( kana_flag = False, pos = 0, cursor_flag = False ;
      *text != '\0' ; text ++, pos ++ ){
    /* ɤȽԤ*/
    kanji_check = is_euc_kanji( *text ) ;
    /* ɽʤΤɤȽꤹ롣*/
    if( pos == node->cur_pos ){
      if( ( y + w->overthespotWin.font_height ) < height ){
	cursor_flag = True ;
      } else if( ( kanji_check &&
		   ( ( x + w->overthespotWin.kanji_width ) < width ) ) ||
		 ( !kanji_check &&
		   ( x + w->overthespotWin.roman_width ) < width ) ){
	cursor_flag = True ;
      }
    }
    /* ɤΰХܤʤΤǤ礦 */
    if( kanji_check && kana_flag == False ){
      kana_flag = True ;
      chara = ( unsigned char )*text ;
      /* ̤αüޤ褷ΤɤȽǤ롣*/
      if( ( x + w->overthespotWin.kanji_width ) >= width ){
	y += w->overthespotWin.font_height ;
	line_num ++ ;
	x = w->overthespotWin.client_area.x ;
	/* ̤βޤǽڤäΤɤȽǤ롣*/
	if( y >= height )
	  goto exit_loop ;
	line_top    = text ;
	linetop_pos = pos ;
	/* 2ܤιƬ֤ФƤ*/
	if( midline_top == NULL ){
	  midline_top = line_top ;
	  midline_pos = linetop_pos ;
	}
      }
    } else {
      /* ɤХܤʤΤǤ礦 */
      if( kana_flag ){
	chara = ( chara << 8 ) | ( ( unsigned char )*text ) ;
	x += w->overthespotWin.kanji_width ;
	kana_flag = False ;
      } else {
	/* ɤäνǤ*/
	if( ( x + w->overthespotWin.roman_width ) >= width ||
	    *text == '\n' ){
	  y += w->overthespotWin.font_height ;
	  line_num ++ ;
	  x = w->overthespotWin.client_area.x ;
	  /* ̤βޤǽڤäΤɤȽǤ롣*/
	  if( y >= height )
	    goto exit_loop ;
	  /* ߤιƬ򵭲Ƥ*/
	  if( *text == '\n' ){
	    line_top    = text + 1 ;
	    linetop_pos = pos + 1 ;
	  } else {
	    line_top    = text ;
	    linetop_pos = pos ;
	  }
	  /* 2ܤιƬ֤ФƤ*/
	  if( midline_top == NULL ){
	    midline_top = line_top ;
	    midline_pos = linetop_pos ;
	  }
	}
	x += w->overthespotWin.roman_width ;
      }
    }
  }
  /* ޤǤǥ뤬ɽƤʤäν*/
  if( !cursor_flag ){
    /* Ԥɬפꡩ */
    if( ( x + w->overthespotWin.roman_width ) >= width ){
      y += w->overthespotWin.font_height ;
      line_num ++ ;
      x = w->overthespotWin.client_area.x ;
      if( y >= height )
	goto exit_loop ;
      line_top    = text ;
      linetop_pos = pos ;
      x += w->overthespotWin.roman_width ;
      cursor_flag = True ;
    } else {
      x += ( w->overthespotWin.roman_width ) ;
      cursor_flag = True ;
    }
  }
 exit_loop:
  /* ɽλ֤κƷ׻*/
  if( x == w->overthespotWin.client_area.x ){
    x = w->overthespotWin.client_area.width +
      w->overthespotWin.client_area.x ;
    y -= w->overthespotWin.font_height ;
    line_num -- ;
  }

  /* ɽ٤Կ򵭲Ƥ*/
  w->overthespotWin.canvas_lines = line_num + 1 ;
  /*
   * ( y - w->overthespotWin.spot_y ) / w->overthespotWin.font_height + 1 ;
   */
  if( !cursor_flag ){
    /* ƥȤɽƤǲ̤ϤߤǤƤޤ⥫ */
    /* Ϥޤо줷ƤʤȤɤ褦̵ˤʤäƤޤ */
    /* ġ*/
    /* ե꡼֤ɤΤ衩 ꡪ
     *   ֤ɤääơĤ뤷ʤ
     */
    OTSW_ConfigureWindowsScreenIsSmall
      ( gw, x, y, line_top, linetop_pos, midline_top, midline_pos ) ;
    return ;
  }

  /* 饦ɥκƷ׻֤ʤɤԤ*/
  if( w->overthespotWin.canvas_lines == 1 ){
    /* ޤȤϻפɡʸʤޤʤäȤ󤸤ʤ *
     * ͤ*/
    if( x <= w->overthespotWin.spot_x ){
      /* ä顢ɥ˴롣*/
      XQueryPopdown( w, topline_probe, topline_popup ) ;
      XQueryPopdown( w, midline_probe, midline_popup ) ;
      XQueryPopdown( w, botline_probe, botline_popup ) ;
      w->overthespotWin.topline_width =
      w->overthespotWin.botline_width = 0 ;
      return ;
    }
    w->overthespotWin.topline_width = x - w->overthespotWin.spot_x ;
  } else {
    w->overthespotWin.topline_width =
      w->overthespotWin.client_area.width - 
	w->overthespotWin.spot_x +
	  w->overthespotWin.client_area.x ;
  }
  w->overthespotWin.topline_text = node->textbuffer ;
  w->overthespotWin.topline_pos  = 0 ;
  w->overthespotWin.topline_x    = w->overthespotWin.spot_x ;

  XtConfigureWidget
    ( w->overthespotWin.topline_popup, 
      w->overthespotWin.topline_x,
      w->overthespotWin.spot_y - w->overthespotWin.font_ascent,
      w->overthespotWin.topline_width,
      w->overthespotWin.font_height, 0 ) ;

  if( !w->overthespotWin.topline_probe ){
    XtPopup( w->overthespotWin.topline_popup, XtGrabNone ) ;
    w->overthespotWin.topline_probe = True ;
  }

  /* ʤν¸ߤν*/
  if( w->overthespotWin.canvas_lines > 2 ){
    w->overthespotWin.midline_text = midline_top ;
    w->overthespotWin.midline_pos  = midline_pos ;

    XtConfigureWidget
      ( w->overthespotWin.midline_popup,
        w->overthespotWin.client_area.x,
        w->overthespotWin.spot_y + w->overthespotWin.font_height - 
        w->overthespotWin.font_ascent,
        w->overthespotWin.client_area.width,
        y - w->overthespotWin.spot_y - w->overthespotWin.font_height,
        0 ) ;
    if( !w->overthespotWin.midline_probe ){
      XtPopup( w->overthespotWin.midline_popup, XtGrabNone ) ;
      w->overthespotWin.midline_probe = True ;
    }
  } else {
    /* õ롣*/
    XQueryPopdown( w, midline_probe, midline_popup ) ;
    w->overthespotWin.midline_text = NULL ;
    w->overthespotWin.midline_pos  = 0 ;
  }

  /* ʤν¸ߤġ*/
  if( w->overthespotWin.canvas_lines > 1 ){
    w->overthespotWin.botline_width =
      x - w->overthespotWin.client_area.x ; 
    w->overthespotWin.botline_text  = line_top ;
    w->overthespotWin.botline_pos   = linetop_pos ;

    /* ɸν*/
    XtConfigureWidget
      ( w->overthespotWin.botline_popup,
        w->overthespotWin.client_area.x,
        y - w->overthespotWin.font_ascent,
        w->overthespotWin.botline_width,
        w->overthespotWin.font_height, 0 ) ;
    if( !w->overthespotWin.botline_probe ){
      XtPopup( w->overthespotWin.botline_popup, XtGrabNone ) ;
      w->overthespotWin.botline_probe = True ;
    }
  } else {
    /* ǲʤĤ롣*/
    XQueryPopdown( w, botline_probe, botline_popup ) ;
    w->overthespotWin.botline_text = NULL ;
    w->overthespotWin.botline_pos  = 0 ;
  }
  return ;

#undef XQueryPopdown
}

/*
 * ⡼ɥɥ֤˰ư硢ȤޤǤɽ
 * ƤޤȤʴؿ
 */
static void OTSW_resetupModeWindow( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  int x, y, min_y, max_y, min_x, max_x ;
  int modeshell_width, modeshell_height ;
  XSetWindowAttributes xswa ;

  modeshell_width  = w->overthespotWin.roman_width * 6 ;
  modeshell_height = w->overthespotWin.font_height ;

  min_x = w->overthespotWin.client_area.x ;
  min_y = w->overthespotWin.client_area.y ;
  max_x = min_x + 
    w->overthespotWin.client_area.width - modeshell_width -
    w->overthespotWin.modeshell_popup->core.border_width ;
  max_y = min_y +
    w->overthespotWin.client_area.height - modeshell_height -
    w->overthespotWin.modeshell_popup->core.border_width ;
  
  xswa.win_gravity = NorthWestGravity ;
  /* ⡼ɥɥư롣*/
  if( !w->overthespotWin.botline_probe &&
      !w->overthespotWin.midline_probe ){
    /* ʸϤȤʸϤƤʤС *
     * Ʊ롣*/
    x = w->overthespotWin.spot_x ;
    y = w->overthespotWin.spot_y -
      w->overthespotWin.font_ascent + w->overthespotWin.font_height ;
  } else {
    x = w->overthespotWin.client_area.x ;
    y = w->overthespotWin.spot_y + 
      w->overthespotWin.font_height * w->overthespotWin.canvas_lines -
      w->overthespotWin.font_ascent ;
  }
  /* ɽ꤬뤫ɤ򸫤롣*/
  if( x > max_x ){
    x = max_x ;
  } else if( x < min_x ){
    x = min_x ;
  }
  /* Τޤɽ̤β˽ФƤޤ */
  if( y > max_y ){
#if 0
    fprintf( stderr, "modeshell y = %d, max_y = %d\n", y, max_y ) ;
    fprintf( stderr, "spot_y = %d, font_ascent = %d, font_height = %d\n",
	     w->overthespotWin.spot_y, w->overthespotWin.font_ascent,
	     w->overthespotWin.font_height ) ;
    fprintf( stderr, "client y = %d, height = %d\n", 
	     w->overthespotWin.client_area.y, 
	     w->overthespotWin.client_area.height ) ;
    fprintf( stderr, "border width = %d\n",
	     w->overthespotWin.modeshell_popup->core.border_width ) ;
#endif
    y = w->overthespotWin.spot_y -
      w->overthespotWin.font_ascent -
      w->overthespotWin.modeshell_popup->core.border_width * 2 -
      w->overthespotWin.font_height ;
    /* ̤ξ夫ФƤޤΡ */
    if( y < min_y ){
      y = w->overthespotWin.spot_y ;
      x = min_x ;
    } else {
      xswa.win_gravity = SouthWestGravity ;
    }
  }
  if( XtIsRealized( gw ) ){
    /* Window  Gravity ꤹ롣*/
    XChangeWindowAttributes
      ( XtDisplay( gw ), XtWindow( w->overthespotWin.modeshell_popup ),
	CWWinGravity, &xswa ) ;
    /* ⡼ɥɥ¸ߤƤʤääƤޤ*/
    if( !w->overthespotWin.modeshell_probe ){
      w->overthespotWin.modeshell_probe = True ;
      XtMoveWidget( w->overthespotWin.modeshell_popup, x, y ) ;
      XtPopup( w->overthespotWin.modeshell_popup, XtGrabNone ) ;
    } else {
      /* ɽ֤ν򤹤롣*/
      XtMoveWidget
	( w->overthespotWin.modeshell_popup, x, y ) ;
      XClearWindow
	( XtDisplay( gw ),
	  XtWindow( w->overthespotWin.modeshell_canvas ) ) ;
    }
    /* ⡼ɥ饤ɽ롣*/
    XtVaSetValues
      ( w->overthespotWin.modeshell_canvas,
        XtNredrawRequest, True, NULL ) ;
  }
  return ;
}

/*
 * skkinput β̤򤤤ľؿ
 *-----
 * ̤礭ѹʤɤԤ줿ˤϤδؿƤФ뤳Ȥ
 * ʤ롣
 */
static void OTSW_RedrawScreen( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;

#define RedrawCanvas(widg,flag)	\
  do {\
    if( w->overthespotWin.flag ){\
      XtVaSetValues\
        (w->overthespotWin.widg, XtNredrawRequest, True, NULL) ;\
    }\
  } while(0)

  /* ɽκƷ׻Ԥ*/
  OTSW_CalcSizeOfBufferWindows( gw ) ;
  /* 褵롣*/
  RedrawCanvas( topline_canvas, topline_probe ) ;
  RedrawCanvas( midline_canvas, midline_probe ) ;
  RedrawCanvas( botline_canvas, botline_probe ) ;
  /* ⡼ɥɥɽ롣*/
  OTSW_resetupModeWindow( gw ) ;
  /* mini-buffer ɽ롣*/
  OTSW_ConfigureMinibuffer( gw ) ;

#undef RedrawCanvas
  return ;
}

/*
 * ɽڤƤޤä(*text,pos)
 */
static void OTSW_ConfigureWindowsScreenIsSmall
( Widget gw,                  int endx, int endy,
  unsigned char *linetop,     int linetop_pos,
  unsigned char *midline_top, int midline_pos )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;
  int width, start_x, toppos ;
  unsigned char *toptext ;

#define XQueryPopdown(w,canvas,popupw) \
do { \
       if( w->overthespotWin.canvas ){ \
			    XtPopdown(w->overthespotWin.popupw) ; \
			    w->overthespotWin.canvas = False ; \
			} \
   } while( 0 )

  /* ǵ줿 pos  text ɽɤ*/
  OTSW_CalcTopPositionOfBuffer
    ( linetop, linetop_pos, node->cur_pos, 
      w->overthespotWin.client_area.width,
      w->overthespotWin.roman_width, w->overthespotWin.kanji_width,
      &toptext, &toppos ) ;

  /* ξ硣ȥåץ饤礭ꤷʤȤʤ*/
  {
    unsigned char *tmp_ptr = toptext ;
    int tmp_pos = toppos, cursor_flag = False ;

    width = 0 ;
    /* ޤɽɤΤ׻롣*/
    while( *tmp_ptr != '\0' &&
	  width < w->overthespotWin.client_area.width ){
      if( tmp_pos == node->cur_pos )
	cursor_flag = True ;
      width += w->overthespotWin.roman_width ;
      tmp_ptr ++ ;
      tmp_pos ++ ;
    }
    if( !cursor_flag ){
      width += w->overthespotWin.roman_width ;
    }
  }
  /* ơɤΥɥưɤΤͤ롣*/
  if( w->overthespotWin.canvas_lines == 1 ){
    if( endx <= w->overthespotWin.spot_x ){
      /* ä顢ɥ˴롣*/
      XQueryPopdown( w, topline_probe, topline_popup ) ;
      XQueryPopdown( w, midline_probe, midline_popup ) ;
      XQueryPopdown( w, botline_probe, botline_popup ) ;
      w->overthespotWin.topline_width =
	w->overthespotWin.botline_width = 0 ;
      return ;
    }
    if( ( width + w->overthespotWin.spot_x ) >=
       ( w->overthespotWin.client_area.x +
	w->overthespotWin.client_area.width ) || linetop_pos != toppos ){
      start_x = w->overthespotWin.client_area.x ;
    } else {
      start_x = w->overthespotWin.spot_x ;
    }
    /* ФʤȤȤϤޤ */
    if( width >= w->overthespotWin.client_area.width ){
      width   = w->overthespotWin.client_area.width ;
    }
    /* ɽϰ֡¾ꡣ*/
    w->overthespotWin.topline_width = width ;
    w->overthespotWin.topline_x     = start_x ;
    w->overthespotWin.topline_text  = toptext ;
    w->overthespotWin.topline_pos   = toppos ;
  } else {
    w->overthespotWin.topline_width =
      w->overthespotWin.client_area.width - w->overthespotWin.spot_x +
	w->overthespotWin.client_area.x ;
    w->overthespotWin.topline_x     = w->overthespotWin.spot_x ;
    w->overthespotWin.topline_text  = node->textbuffer ;
    w->overthespotWin.topline_pos   = 0 ;
  }
  XtConfigureWidget
    ( w->overthespotWin.topline_popup,
      w->overthespotWin.topline_x,
      w->overthespotWin.spot_y - w->overthespotWin.font_ascent,
      w->overthespotWin.topline_width,
      w->overthespotWin.font_height, 0 ) ;
  /* ǾʤΥ饤ɽpopup롣*/
  if( !w->overthespotWin.topline_probe ){
    XtPopup( w->overthespotWin.topline_popup,XtGrabNone ) ;
    w->overthespotWin.topline_probe = True ;
  }
  /* ʤν¸ߤν*/
  if( w->overthespotWin.canvas_lines > 2 ){
    w->overthespotWin.midline_text = midline_top ;
    w->overthespotWin.midline_pos  = midline_pos ;
    XtConfigureWidget
      ( w->overthespotWin.midline_popup,
        w->overthespotWin.client_area.x,
        w->overthespotWin.spot_y + 
	w->overthespotWin.font_height - w->overthespotWin.font_ascent,
        w->overthespotWin.client_area.width,
        endy - w->overthespotWin.spot_y -
        w->overthespotWin.font_height,  0 ) ;
    if( !w->overthespotWin.midline_probe ){
      XtPopup( w->overthespotWin.midline_popup, XtGrabNone ) ;
      w->overthespotWin.midline_probe = True ;
    }
  } else {
    XQueryPopdown( w, midline_probe, midline_popup ) ;
    w->overthespotWin.midline_text = NULL ;
    w->overthespotWin.midline_pos  = 0 ;
  }
  /* ʤν¸ߤġ*/
  if( w->overthespotWin.canvas_lines > 1 ){
    if( width >= w->overthespotWin.client_area.width ){
      width   = w->overthespotWin.client_area.width ;
    }
    w->overthespotWin.botline_width = width ;
    w->overthespotWin.botline_text  = toptext ;
    w->overthespotWin.botline_pos   = toppos ;

    XtConfigureWidget
      ( w->overthespotWin.botline_popup,
        w->overthespotWin.client_area.x, 
        endy - w->overthespotWin.font_ascent,
        w->overthespotWin.botline_width,
        w->overthespotWin.font_height, 0 ) ;
    if( !w->overthespotWin.botline_probe ){
      XtPopup( w->overthespotWin.botline_popup, XtGrabNone ) ;
      w->overthespotWin.botline_probe = True ;
    }
  } else {
    /* ǲʤĤ롣*/
    XQueryPopdown( w, botline_probe, botline_popup ) ;
    w->overthespotWin.botline_text = NULL ;
  }
  return ;

#undef XQueryPopdown
}

/* ǡΰ֤̳˽ФΤɤȽǤʤФʤ *
 * ʤ*/
static void OTSW_ConfigureMinibufferPosition( Widget gw, int *rx, int *ry )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  XWindowAttributes xwa ;
  unsigned int mheight, mwidth ;
  int y_lower, y_base, y, x ;
  int win_gravity ;
  XSizeHints sizehints ;

  win_gravity = NorthWestGravity ;

  /* ɽ륦ɥΰ֤롣ϵִ֤ *
   * 롣Ͼ˾ꤷƤ͡*/
  y_base = w->overthespotWin.client_y + 
    w->overthespotWin.focus_offset_y + 
    w->overthespotWin.spot_y - w->overthespotWin.font_ascent ;
  x = w->overthespotWin.client_x + w->overthespotWin.focus_offset_x ;

  if( !w->overthespotWin.topline_probe ){
    y = y_base + w->overthespotWin.font_height ;
  } else {
    /* ǽΰԲɽƤˤϡĥХ˽񤫤Ƥ *
     * ΤäƤޤʤ褦֤롣*/
    y = y_base + w->overthespotWin.font_height *
      w->overthespotWin.canvas_lines ;
  }
  /* 롼ȥɥ礭롣*/
  XGetWindowAttributes
    ( XtDisplay( gw ),
      RootWindowOfScreen
      ( XtScreen( w->overthespotWin.minibuffer_popup ) ),
      &xwa ) ;
#ifdef DEBUG
  fprintf( stderr, "Root Window Height : %d\n", xwa.height ) ;
#endif
  /* ߥ˥Хåեι⤵ФƤ٤ */
  mheight = w->overthespotWin.minibuffer_popup->core.border_width * 2 +
    ( ( w->overthespotWin.change_minibuffer_font )?
      w->overthespotWin.font_height : w->overthespotWin.mfont_height ) ;
  mwidth = w->overthespotWin.minibuffer_popup->core.border_width * 2 +
    ( ( w->overthespotWin.change_minibuffer_font )?
      w->overthespotWin.roman_width : w->overthespotWin.mroman_width ) *
    w->overthespotWin.minibuf_width ;

  /* κɸߥ˥ХåեȤΰֲκɸˤʤ롣*/
  y_lower = y + mheight ;
#ifdef DEBUG
  fprintf( stderr, "Minibuffer : y (%d) + height (%d)\n", 
	   y, mheight ) ;
#endif
  /* ̳˽ФƤޤΡ */
  if( y_lower > xwa.height ){
    if( y_base > xwa.height ){
      /* ʤ̤ΰֲ˶ŪˤϤäĤޤ*/
      y = xwa.height - mheight -
	w->overthespotWin.minibuffer_popup->core.border_width * 2 ;
    } else {
      /* ʤ饫ξɽޤ礦*/
      y = y_base - mheight -
	w->overthespotWin.minibuffer_popup->core.border_width * 2 ;
    }
    win_gravity = SouthWestGravity ;
  }
  /* ⤷Ʋ̤ξ˽ФƤޤȤȤΤǤ */
  if( y < 0 ){
    /* ʤΤޤ󡣤⤦̤ξ֤Ƥޤޤ*/
    y = 0 ;
    win_gravity = NorthWestGravity ;
  }
#ifdef DEBUG
  fprintf( stderr, "Result Y = %d, y_base = %d\n", y, y_base ) ;
#endif
  if( x + mwidth > xwa.width ){
    x = xwa.width - mwidth ;
    if( x < 0 ){
      x = 0 ;
    } else {
      if( win_gravity == NorthWestGravity ){
	win_gravity = NorthEastGravity ;
      } else {
	win_gravity = SouthEastGravity ;
      }
    }
  } else if( x < 0 ){
    x = 0 ;
  }
  /* Window Manager ˥ҥȤäƤ롣XSetWMNormalHints * 
   * ʬ񴹤ʤ褦ʤΤǡ hints ɤߤȤäƤ麹*
   * ؤȤˡȤʤФʤʤ*/
  sizehints.flags = PWinGravity ;
  sizehints.win_gravity = win_gravity ;
  XChangeWMNormalHints
    ( XtDisplay( gw ),
      XtWindow( w->overthespotWin.minibuffer_popup ),
      &sizehints ) ;
  *rx = x ;
  *ry = y ;
  return  ;
}

/*
 * ߥ˥Хåեɥ֡礭Ԥؿ
 *----
 * 礭ϼ¤ꤷƤޤ󡣺ǽꤷǤ̤˥ߥ
 * Хåե򲿽֤ɤΤʡĤȽǤ˻Ȥ櫓Ǥ
 */
static void OTSW_ConfigureMinibuffer( Widget gw )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )gw ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;
  struct SKKInputNode *mNode = w->overthespotWin.lastbuffer ;

  if( mNode == node ){
    /* ɽ٤ƥȤ̵äˤϡ*/
    if( node->mtextbuffer[ 0 ] == '\0' ){
      /* УϣХåפ֤ä顢Ĥ롣*/ 
      if( w->overthespotWin.minibuffer_probe ){
	XtPopdown( w->overthespotWin.minibuffer_popup ) ;
	w->overthespotWin.minibuffer_probe = False ;
      }
      return ;
    }
  }
  /* ߥ˥ХåեϵƤʤΡ ƤʤΤä鵯ɬ *
   * 衣*/
  if( !w->overthespotWin.minibuffer_probe ){
    int x, y ;
    /* ꤵ줿֤˥ߥ˥Хåեư롣*/
    OTSW_ConfigureMinibufferPosition( gw, &x, &y ) ;
    XtMoveWidget
      ( w->overthespotWin.minibuffer_popup, x, y ) ;
    /* Popup 롣*/
    XtPopup( w->overthespotWin.minibuffer_popup, XtGrabNone ) ;
    w->overthespotWin.minibuffer_probe = True ;
  }
  /* Хľ褦׵Ф*/
  XtVaSetValues
    ( w->overthespotWin.minibuffer_canvas, 
      XtNredrawRequest, True, NULL ) ;
  return ;
}    

/*
 * ʲcanvasWidget Ѥ callback ã
 */
static void Topline_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )client ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;

  if( !w->overthespotWin.topline_probe )
    return ;

  XClearWindow
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.topline_canvas ) ) ;
  OTSW_FullDrawTopBufferTopLine
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.topline_canvas ),
      w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
      w->overthespotWin.roman_rgc, w->overthespotWin.kanji_rgc,
      w->overthespotWin.topline_text,
      w->overthespotWin.topline_pos,
      w->overthespotWin.roman_width, w->overthespotWin.kanji_width,
      w->overthespotWin.font_ascent, w->overthespotWin.font_height,
      w->overthespotWin.topline_width,
      ( ( w->overthespotWin.botline_probe ||
	  w->overthespotWin.midline_probe ||
	( w->overthespotWin.topline_x +
	  w->overthespotWin.topline_width ) >=
	  w->overthespotWin.client_area.x + 
	  w->overthespotWin.client_area.width )? True : False ),
      node->cur_pos, node->cur_exist, w->overthespotWin.is_focus ) ;
  return ;
}

static void Midline_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )client ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;

  if( !w->overthespotWin.midline_probe )
    return ;
    
  XClearWindow
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.midline_canvas ) ) ;
  /* ιԤ褹롣*/
  OTSW_FullDrawTopBufferMiddleLines
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.midline_canvas ),
      w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
      w->overthespotWin.roman_rgc, w->overthespotWin.kanji_rgc,
      w->overthespotWin.midline_text,
      w->overthespotWin.midline_pos,
      w->overthespotWin.roman_width, w->overthespotWin.kanji_width,
      w->overthespotWin.font_ascent, w->overthespotWin.font_height,
      w->overthespotWin.client_area.width,
      w->overthespotWin.canvas_lines - 2,
      node->cur_pos, node->cur_exist, w->overthespotWin.is_focus ) ;
  return ;
}

static void Botline_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )client ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;

  if( !w->overthespotWin.botline_probe )
    return ;

  XClearWindow
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.botline_canvas ) ) ;
  /* ֲιԤ褹롣*/
  OTSW_FullDrawTopBufferTopLine
    ( XtDisplay( gw ), XtWindow( w->overthespotWin.botline_canvas ),
      w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
      w->overthespotWin.roman_rgc, w->overthespotWin.kanji_rgc,
      w->overthespotWin.botline_text,
      w->overthespotWin.botline_pos,
      w->overthespotWin.roman_width, w->overthespotWin.kanji_width,
      w->overthespotWin.font_ascent, w->overthespotWin.font_height,
      w->overthespotWin.botline_width, 
      ( w->overthespotWin.botline_width <
        w->overthespotWin.client_area.width )? False : True,
      node->cur_pos, node->cur_exist, w->overthespotWin.is_focus ) ;
  return ;
}

/*
 * δؿϸߤξǤκ褷ޤpopup/popdown ʤɤ
 * ƤФäƤƤ뤳Ȥˤޤꥢ꤫ʤ Ǥ⡢
 * ѲڤäưͿΤϿ(Ĥޤ overthespotWinWidget)ʤΤ
 * 顢ҶǤȤ餻äƤΤϡġo
 */
static void Minibuffer_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )client ;
  struct SKKInputNode *mNode ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;

  if( !w->overthespotWin.minibuffer_probe )
    return ;

  XClearWindow( XtDisplay( gw ), XtWindow( gw ) ) ;

  /* mini-buffer ɽ롣*/
  mNode = w->overthespotWin.lastbuffer ;

  if( mNode == node ){
    /* ɽ٤ƥȤ̵äˤϡ*/
    if( node->mtextbuffer[ 0 ] == '\0' )
      return ;

    /* ĤߤäȤʤɬɽ˽ФƤ褦ˤ롣*/
    XRaiseWindow
      ( XtDisplay( gw ), XtWindow( gw ) ) ;
    if( w->overthespotWin.change_minibuffer_font ){
      /* ߥ˥Хåեɽ롣*/
      skkwin_jputstring
	( XtDisplay( gw ), XtWindow( gw ), 
	  w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
	  0, w->overthespotWin.font_ascent, node->mtextbuffer,
	  w->overthespotWin.roman_width, w->overthespotWin.kanji_width ) ;
    } else {
      skkwin_jputstring
	( XtDisplay( gw ), XtWindow( gw ), 
	  w->overthespotWin.mroman_ngc, w->overthespotWin.mkanji_ngc,
	  0, w->overthespotWin.mfont_ascent, node->mtextbuffer,
	  w->overthespotWin.mroman_width, w->overthespotWin.mkanji_width ) ;
    }
  } else {
    /* ĤߤäȤʤɬɽ˽ФƤ褦ˤ롣*/
    XRaiseWindow
      ( XtDisplay( gw ), XtWindow( gw ) ) ;

    if( mNode->mtextbuffer[ 0 ] != '\0' ){
      if( w->overthespotWin.change_minibuffer_font ){
	skkwin_jputstring
	  ( XtDisplay( gw ), XtWindow( gw ),
	    w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
	    0, w->overthespotWin.font_ascent, mNode->mtextbuffer,
	    w->overthespotWin.roman_width,
	    w->overthespotWin.kanji_width ) ;
      } else {
	skkwin_jputstring
	  ( XtDisplay( gw ), XtWindow( gw ),
	    w->overthespotWin.mroman_ngc, w->overthespotWin.mkanji_ngc,
	    0, w->overthespotWin.mfont_ascent, mNode->mtextbuffer,
	    w->overthespotWin.mroman_width,
	    w->overthespotWin.mkanji_width ) ;
      }
    } else {
      unsigned char *toptext ;
      int toppos ;
      XWindowAttributes xwa ;
      /* ǲˤʤ뤫ĲȴФƤ*/
      XGetWindowAttributes
	( XtDisplay( gw ), XtWindow( gw ), &xwa ) ;

      if( w->overthespotWin.change_minibuffer_font ){
	/* Window  buffer ʸɽ롣*/
	OTSW_CalcTopPositionOfBuffer
	  ( mNode->textbuffer, 0, mNode->cur_pos, xwa.width,
	    w->overthespotWin.roman_width,
	    w->overthespotWin.kanji_width,
	    &toptext, &toppos ) ;
	OTSW_FullDrawTopBufferTopLine
	  ( XtDisplay( gw ), XtWindow( gw ),
	    w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
	    w->overthespotWin.roman_rgc, w->overthespotWin.kanji_rgc,
	    toptext, toppos,
	    w->overthespotWin.roman_width, w->overthespotWin.kanji_width,
	    w->overthespotWin.font_ascent, w->overthespotWin.font_height,
	    xwa.width, True,
	    mNode->cur_pos, mNode->cur_exist, 
	    w->overthespotWin.is_focus ) ;
      } else {
	/* Window  buffer ʸɽ롣*/
	OTSW_CalcTopPositionOfBuffer
	  ( mNode->textbuffer, 0, mNode->cur_pos, xwa.width,
	    w->overthespotWin.mroman_width,
	    w->overthespotWin.mkanji_width,
	    &toptext, &toppos ) ;
	OTSW_FullDrawTopBufferTopLine
	  ( XtDisplay( gw ), XtWindow( gw ),
	    w->overthespotWin.mroman_ngc, w->overthespotWin.mkanji_ngc,
	    w->overthespotWin.mroman_rgc, w->overthespotWin.mkanji_rgc,
	    toptext, toppos,
	    w->overthespotWin.mroman_width, w->overthespotWin.mkanji_width,
	    w->overthespotWin.mfont_ascent, w->overthespotWin.mfont_height,
	    xwa.width, True,
	    mNode->cur_pos, mNode->cur_exist, 
	    w->overthespotWin.is_focus ) ;
      }
    }
  }
  return ;
}

static void Modeshell_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )client ;
  struct SKKInputNode *node = w->overthespotWin.topbuffer ;

  /* ⡼ɥɥ¸ߤƤʤääƤޤ*/
  if( !w->overthespotWin.modeshell_probe )
    return ;
  /* ľΤ˾õƤ*/
  XClearWindow( XtDisplay( gw ), XtWindow( gw ) ) ;
  /* ⡼ɥ饤ɽ롣ʬɽľбͽǤϤ롣*/
  OTSW_FullDrawModeLine
    ( XtDisplay( gw ), XtWindow( gw ),
      w->overthespotWin.roman_ngc, w->overthespotWin.kanji_ngc,
      node, 0, w->overthespotWin.font_ascent,
      w->overthespotWin.roman_width, w->overthespotWin.kanji_width,
      w->overthespotWin.jisyo_dirty ) ;
  return ;
}

/*
 * Canvas Widget ˸ƤӽФ륳Хåؿ
 */
static void Canvas_DestroyCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  XtRemoveAllCallbacks( gw, XtNredrawNotify ) ;
  XtRemoveAllCallbacks( gw, XtNdestroyNotify ) ;
  return ;
}

static void OTSW_wmMessageCloseCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  OverthespotWinWidget w = ( OverthespotWinWidget )client ;
  struct SKKInputNode *node ;
  unsigned char keybuf[ 3 ] ;
  int keybuf_use ;

  node = ( w->overthespotWin.lastbuffer->cur_exist )?
    w->overthespotWin.lastbuffer : w->overthespotWin.topbuffer ;

  strcpy( keybuf, w->overthespotWin.keybuf ) ;
  keybuf_use = w->overthespotWin.keybuf_use ;
  
  w->overthespotWin.keybuf[ 0 ] = 0x07 ;
  w->overthespotWin.keybuf[ 1 ] = '\0' ;
  w->overthespotWin.keybuf_use  = 1 ;

  if( node->j_henkan_show_candidate_mode ){
    do_Function_jhenkanShowCandidateMode
      ( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	node, w->overthespotWin.keybuf ) ;
  } else if( node->j_input_by_code_or_menu_mode ){
    if( node->j_input_by_code_or_menu_mode == True ){
      do_function_j_input_by_code_or_menu_jump
	( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	  node, w->overthespotWin.keybuf ) ;
    } else {
      do_function_j_input_by_code_or_menu_1_jump
	( gw, &w->overthespotWin.topbuffer, &w->overthespotWin.lastbuffer,
	  node, w->overthespotWin.keybuf ) ;
    }
  } else if( w->overthespotWin.topbuffer !=
	     w->overthespotWin.lastbuffer ){
    j_keyboard_quit
      ( gw, &w->overthespotWin.topbuffer,
	&w->overthespotWin.lastbuffer, node ) ;
  } else {
    node->mtextbuffer[ 0 ] = '\0' ;
  }
  strcpy( w->overthespotWin.keybuf, keybuf ) ;
  w->overthespotWin.keybuf_use = keybuf_use ;
  OTSW_RedrawScreen( ( Widget )w ) ;
  return ;
}

static void OTSW_wmShellDestroyCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  /* ƤΥХåؿ˴Ƥ*/
  XtRemoveAllCallbacks( gw, XtNwmcloseNotify ) ;
  XtRemoveAllCallbacks( gw, XtNdestroyNotify ) ;
  return ;
}
