/* # skkinput (Simple Kana-Kanji Input)
 * SkkInput.c --- Making SeparateWidget.
 * This file is part of skkinput.
 * Copyright (C) 1997
 * Takashi SAKAMOTO (sakamoto@yajima.kuis.kyoto-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "SeparateWinP.h"
#include "config.h"
#include "skkkey.h"
#include "otwexfunc.h"
#include "Canvas.h"
#include "WmcloseShell.h"

#ifndef XtNgeometry
#define XtNgeometry		"geometry"
#define XtCGeometry		"Geometry"
#endif

#define offset(field) XtOffsetOf( SeparateWinRec, separateWin.field)
#define goffset(field) XtOffsetOf( WidgetRec, core.field )

static XtResource resources[] = {
  /* Ȥ餢꥽*/
  { XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
    goffset(width), XtRImmediate, (XtPointer) 80},
  { XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
    goffset(height), XtRImmediate, (XtPointer) 3},
  { XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
    goffset(background_pixel), XtRString, XtDefaultBackground },
  /* ʬȤǺä꥽*/
  { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
    offset(puppixel), XtRString, XtDefaultForeground },
  { XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
    offset(reverse_video), XtRImmediate, (XtPointer) FALSE},
  { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    offset(fs_roman), XtRString, DEFAULT_ROMANFONT },
  { XtNkanjiFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    offset(fs_kanji), XtRString, DEFAULT_KANJIFONT },
  /* skkinput 뤬ĤѤ callback */
  { XtNendNotify, XtCCallback, XtRCallback, sizeof(caddr_t), 
    offset(endcallback), XtRCallback, (caddr_t)NULL },
  /* skkinput  client ʸѤ callback */
  { XtNfixNotify, XtCCallback, XtRCallback, sizeof(caddr_t), 
    offset(fixcallback), XtRCallback, (caddr_t)NULL },
  /* key event  client ֤Ѥ callback */
  { XtNkeybackNotify, XtCCallback, XtRCallback, sizeof(caddr_t), 
    offset(keybackcallback), XtRCallback, (caddr_t)NULL },
  /* Shift-Space  skkinput Ĥ褦ˤ뤫ݤ*/
  { XtNcompatibleCloseSkkinputKey, XtCCompatibleCloseSkkinputKey,
    XtRBoolean, sizeof ( Boolean ),
    offset(compatible_close_skkinputkey), XtRImmediate, (XtPointer)TRUE },
  /* egg ߴ j-newline ݤ*/
  { XtNeggLikeNewline, XtCEggLikeNewline, XtRImmediate, sizeof (int),
    offset(egg_like_newline), XtRImmediate, (XtPointer) FALSE},
  /* chat-adapter-mode ݤ*/
  { XtNchatAdapter, XtCChatAdapter, XtRImmediate, sizeof (int),
    offset(chat_adapter), XtRImmediate, (XtPointer) FALSE},
  { XtNjisyoDirty, XtCJisyoDirty, XtRImmediate, sizeof (int),
    offset(jisyo_dirty), XtRImmediate, (XtPointer) FALSE },
  { XtNclearMinibuffer, XtCClearMinibuffer, XtRImmediate, sizeof( int ),
    offset( clearMinibuffer ), XtRImmediate, (XtPointer)FALSE },
  /* history νѤؿ*/
  { XtNconversionHistory, XtCConversionHistory, XtRImmediate,
    sizeof( HistoryListNode * ), offset( historyAttribute ),
    XtRImmediate, ( XtPointer )NULL },
  { XtNsetFocus, XtCSetFocus, XtRImmediate, sizeof( int ),
    offset( conversion_set_focus ), XtRImmediate, (XtPointer)FALSE },
  { XtNunsetFocus, XtCUnsetFocus, XtRImmediate, sizeof( int ),
    offset( conversion_unset_focus ), XtRImmediate, (XtPointer)FALSE },
  /* Subwindow ե֤ɥꤹ롣*/
  { XtNprobeWindow, XtCProbeWindow, XtRImmediate, sizeof( Window ),
    offset( probe_window ), XtRImmediate, ( XtPointer )None },
} ;

#undef offset
#undef goffset

/*
 * ץȥ
 */
/*
 * ΥåȤƤǤȤʤؿ
 */
static void SEPW_Initialize
( Widget greq, Widget gnew, ArgList args, Cardinal *num_args ) ;
static void SEPW_Realize
( Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs ) ;
static void SEPW_Redisplay
( Widget gw, XEvent *event, Region region ) ;
static void SEPW_Destroy( Widget gw ) ;
static Boolean SEPW_SetValues
( Widget current, Widget request, Widget new,
  ArgList args, Cardinal *num_args ) ;
static void KeyDownEventHandler
( Widget gw, XEvent *event, String *params, Cardinal *num_params ) ;
static void FocusInEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params ) ;
static void FocusOutEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params ) ;

static int do_function_by_reading_character
( Widget gw, struct SKKInputNode *node ) ;
static void SEPW_RedrawScreen( Widget gw ) ;
static void skkinput_DrawScreen( Widget gw ) ;
static void SEPW_ClearMinibuffer( Widget gw ) ;

static void SeparateCanvas_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void Canvas_DestroyCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void WMMessageCloseCallback
( Widget gw, caddr_t client, caddr_t caller ) ;
static void WMShellDestroyCallback
( Widget gw, caddr_t client, caddr_t caller ) ;

/*
 * ȤƤؿΥץȥ
 */
extern void skkwin_jerase
( Display *disp, Window win, GC gc, int x, int y,
  int font_ascent, int width, int height ) ;
extern void skkwin_jputchar
( Display *disp, Window win, GC roman_gc, GC kanji_gc,
  int x, int y, int chara, int flag ) ;
extern void skkwin_jputchar2
( Display *disp, Window win,
  GC roman_ngc, GC kanji_ngc, GC roman_rgc, GC kanji_rgc,
  int x, int y, int chara, 
  int roman_width, int kanji_width, int font_height, 
  int font_ascent, int flag ) ;
extern void skkwin_jputstring
( Display *disp, Window win, GC roman_gc, GC kanji_gc,
  int x, int y, char *text, int roman_width, int kanji_width ) ;
extern void skkwin_jdiffputstring
( Display *disp, Window win, 
  GC roman_ngc, GC kanji_ngc, GC roman_rgc, GC kanji_rgc,
  int x, int y, int width,
  char *text, char *ptext, int roman_width, int kanji_width,
  int font_height, int font_ascent ) ;
extern void skkinputDrawCursor
( Display *disp, Window win, GC roman_ngc, GC roman_rgc, 
  int x, int y, int font_ascent, int width, int height,
  int is_focus ) ;
extern int skkinput_GetLineNum
( unsigned char *text, int roman_width, int kanji_width,
  unsigned int width, int cur_pos, int *cur_line ) ;
extern int skkinput_FindStartLine
( unsigned char *text, int roman_width, int kanji_width,
  int startline, unsigned int width ) ;

/* MyDispatch.c */
extern int add_myeventhandler
( Display *disp, Window src_win, Window dest_win,
  int event_type, long event_mask ) ;
extern int remove_allmyeventhandler
( Display *disp, Window src_win ) ;

/* MyError.c */
extern void ignore_xerror( void ) ;
extern void regard_xerror( void ) ;

/*
 * Хѿ
 */
static XtActionsRec separateWinActionsTable [] = {
  { "KeyDownEventHandler",	   KeyDownEventHandler },
  { "FocusInEventHandler",	   FocusInEventHandler },
  { "FocusOutEventHandler",	   FocusOutEventHandler },
} ;

static char defaultSeparateWinTranslations[] =  
"<Key>:                   KeyDownEventHandler()\n\
 <FocusIn>:               FocusInEventHandler()\n\
 <FocusOut>:              FocusOutEventHandler()\n" ;

SeparateWinClassRec separateWinClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Separate",
    /* size			*/	sizeof(SeparateWinRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	SEPW_Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	SEPW_Realize,
    /* actions			*/	separateWinActionsTable,
    /* num_actions		*/	XtNumber( separateWinActionsTable ),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber( resources ),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	SEPW_Destroy,
    /* resize			*/	NULL,
    /* expose			*/	SEPW_Redisplay,
    /* set_values		*/	SEPW_SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	defaultSeparateWinTranslations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    }
};

WidgetClass separateWinWidgetClass = ( WidgetClass )&separateWinClassRec ;

static void SEPW_Initialize 
( Widget greq, Widget gnew, ArgList args, Cardinal *num_args )
{
  SeparateWinWidget w = ( SeparateWinWidget )gnew ;
  struct SKKInputNode *node ;
  XtGCMask	valuemask ;
  XGCValues	myXGCV ;
  int roman_width, roman_ascent, roman_height ;
  int kanji_width, kanji_ascent, kanji_height ;
  int fontwidth_sub1, fontwidth_sub2 ;
  
  /* եȤξɤ߽ФƤ*/
  roman_width  = w->separateWin.fs_roman->max_bounds.width ;
  roman_ascent = w->separateWin.fs_roman->ascent ;
  roman_height = roman_ascent + w->separateWin.fs_roman->descent ;
  kanji_width  = w->separateWin.fs_kanji->max_bounds.width ;
  kanji_ascent = w->separateWin.fs_kanji->ascent ;
  kanji_height = kanji_ascent + w->separateWin.fs_kanji->descent ;

  fontwidth_sub1 = ( roman_width << 1 ) ;
  fontwidth_sub2 = ( kanji_width >> 1 ) ;
  /* ߤΥեȤ礭δطѤʤΤ⤷ʤ*/
  if( fontwidth_sub1 != kanji_width ){
    /* ѿܤΤ礭 */
    if( fontwidth_sub1 > kanji_width ){
      kanji_width = fontwidth_sub1 ;
    } else {
      /* ѿܤΤϴϾ餷*/
      roman_width = fontwidth_sub2 ;
    }
  }
  w->separateWin.roman_width = roman_width ;
  w->separateWin.kanji_width = kanji_width ;

  w->separateWin.font_height = ( ( kanji_height > roman_height )?
    kanji_height : roman_height ) ;
  w->separateWin.font_ascent = ( ( kanji_ascent > roman_ascent )?
    kanji_ascent : roman_ascent ) ;

  /* ե٥åɽΤ GC 롣*/
  myXGCV.foreground = w->separateWin.puppixel ;
  myXGCV.background = w->core.background_pixel ;
  myXGCV.font       = w->separateWin.fs_roman->fid ;
  valuemask = GCForeground | GCBackground | GCFont ;
  w->separateWin.roman_ngc = XtGetGC( gnew, valuemask, &myXGCV ) ;

  myXGCV.foreground = w->core.background_pixel ;
  myXGCV.background = w->separateWin.puppixel ;
  myXGCV.font       = w->separateWin.fs_roman->fid ;
  valuemask = GCForeground | GCBackground | GCFont ;
  w->separateWin.roman_rgc = XtGetGC( gnew, valuemask, &myXGCV ) ;

  /* ʸ(Ҳ̾ʿ̾)Τ GC 롣*/
  myXGCV.foreground = w->separateWin.puppixel ;
  myXGCV.background = w->core.background_pixel ;
  myXGCV.font       = w->separateWin.fs_kanji->fid ;
  valuemask = GCForeground | GCBackground | GCFont ;
  w->separateWin.kanji_ngc = XtGetGC( gnew, valuemask, &myXGCV ) ;

  myXGCV.foreground = w->core.background_pixel ;
  myXGCV.background = w->separateWin.puppixel ;
  myXGCV.font       = w->separateWin.fs_kanji->fid ;
  valuemask = GCForeground | GCBackground | GCFont ;
  w->separateWin.kanji_rgc = XtGetGC( gnew, valuemask, &myXGCV ) ;

  /* ¾ν*/
  w->separateWin.pwrite_string[ 0 ] = '\0' ;
  w->separateWin.pmtextbuffer[ 0 ]  = '\0' ;
  w->separateWin.cutbuffer[ 0 ]  = '\0' ;
  w->separateWin.keybuf_use    = 0 ;
  w->separateWin.keybuf[ 0 ]   = '\0' ;
  w->separateWin.prev_modeline = 0 ;

  /* ҥȥϥꥢƤɸνɬפǤĤޤϰ */
  /* ⵯ줿ȤʤƤʤΤǴǤ롣*/
  w->separateWin.historybuffer = NULL ;
  w->separateWin.hist_start = w->separateWin.hist_end = 0 ;
  w->separateWin.hist_cur   = -1 ;
 
  /* ԽѤΥХåեγݡ*/
  node = skkinput_AllocateMinibuffer() ;
  node->cur_exist = True ;
  w->separateWin.topbuffer  = node ;
  w->separateWin.lastbuffer = node ;

  w->separateWin.prev_jisyo_dirty  = w->separateWin.jisyo_dirty = 0 ;
  w->separateWin.prev_chat_adapter = w->separateWin.chat_adapter ;
  w->separateWin.prev_eggnl        = w->separateWin.egg_like_newline ;

  /* ޤFocus ϺǽϳƤȻפ衣*/
  w->separateWin.is_focus = False ;
  w->separateWin.canvas_probe = False ;

  w->separateWin.historyAttribute = NULL ;
  return ;
}

static void SEPW_CanvasPopupPositionInit
( Widget gw, Window client, 
  int canvas_width, int canvas_height, int canvas_border )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  Window win ;
  int x, y, pos_x, pos_y ;
  unsigned int client_width, client_height, border_width, depth ;
  unsigned int root_width, root_height ;
  long supplied ;
  XSizeHints sizehints ;
  
  XGetGeometry
    ( XtDisplay( gw ), client, &win, &x, &y,
      &client_width, &client_height,
      &border_width, &depth ) ;
  XGetGeometry
    ( XtDisplay( gw ), RootWindowOfScreen( XtScreen( gw ) ),
      &win, &x, &y, &root_width, &root_height, &border_width, &depth ) ;
  XTranslateCoordinates
    ( XtDisplay( gw ), client, RootWindowOfScreen( XtScreen( gw ) ),
      0, 0, &x, &y, &win ) ;
  pos_x = x ;
  pos_y = y + client_height ;
  if( ( pos_y + canvas_height ) > root_height ){
    pos_y = y - canvas_height - canvas_border ;
    if( pos_y < 0 ){
      pos_y = root_height - canvas_height - canvas_border ;
    }
    /* Window Manager ˥ҥȤäƤ롣XSetWMNormalHints * 
     * ʬ񴹤ʤ褦ʤΤǡ hints ɤߤȤäƤ麹*
     * ؤȤˡȤʤФʤʤ*/
    XGetWMNormalHints
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas_popup ),
 	&sizehints, &supplied ) ;
    sizehints.flags |= PWinGravity ;
    sizehints.win_gravity = SouthWestGravity ;
    /* Window Manager ˥ҥȤäƤ롣*/
    XSetWMNormalHints
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas_popup ),
 	&sizehints ) ;
  }
  if( pos_x + canvas_width > root_width ){
    pos_x = root_width - canvas_width - canvas_border ;
  }
  if( pos_x < 0 )
    pos_x = 0 ;
  XtMoveWidget( w->separateWin.canvas_popup, pos_x, pos_y ) ;
  return ;
}

/*
 * Widget ΤĻ˸ƤФؿ
 *------
 * create ǤϼΤʤcreate  realize 뤳
 * ˤäƼΤġλ˸ƤФؿ
 */
static void SEPW_Realize
( Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  Display *display    = XtDisplay( gw ) ;
  /* 줬ʤɡĤ*/
  unsigned int winwidth, winheight ; 
  XSizeHints sizehints ;
  Widget canvas_popup, canvas ;
  CoreWidgetClass super =
    ( CoreWidgetClass )XtClass( gw )->core_class.superclass ;

  /* ŬڤʥɥѰդƤ*/
  ( *super->core_class.realize )( gw, valueMask, attrs ) ;

  winwidth  = w->core.width  * w->separateWin.roman_width ;
  winheight = w->core.height * w->separateWin.font_height ;
  
  /* ޤݥåץåץ롣*/
  canvas_popup = XtVaCreatePopupShell
    ( "skkinput", wmcloseShellWidgetClass, gw, XtNinput, TRUE, 
      XtNmappedWhenManaged, TRUE, 
      XtNwidth, winwidth, XtNheight, winheight, NULL ) ;
  XtAddCallback
    ( canvas_popup, XtNwmcloseNotify, 
      ( XtCallbackProc )WMMessageCloseCallback, gw ) ;
  XtAddCallback
    ( canvas_popup, XtNdestroyNotify, 
      ( XtCallbackProc )WMShellDestroyCallback, gw ) ;
  /* βˤĤɽѤ widget 롣shellwidget ľ *
   * ܽ񤤤ƤϤޤäȤΤǡġ*/
  canvas = XtVaCreateManagedWidget
    ( "canvas", canvasWidgetClass, canvas_popup, NULL ) ;
  XtAddCallback
    ( canvas, XtNredrawNotify, 
      ( XtCallbackProc )SeparateCanvas_RedisplayCallback, gw ) ;
  XtAddCallback
    ( canvas, XtNdestroyNotify, 
      ( XtCallbackProc )Canvas_DestroyCallback, gw ) ;
  /* realize Ƥ*/
  w->separateWin.canvas_popup = canvas_popup ;
  w->separateWin.canvas = canvas ;
  XtRealizeWidget( canvas_popup ) ;
  XtRealizeWidget( canvas ) ;

  /* ηǤ򲿤ȤǤ褦ˤɤȤϻפ衣*/
  winwidth  += canvas_popup->core.border_width * 2 ;
  winheight += canvas_popup->core.border_width * 2 ;

  /* Ƥ˥ꥵäƤ롩 */
  (void) XtMakeResizeRequest
    ( ( Widget )canvas_popup,
      ( Dimension )winwidth, ( Dimension )winheight,
      &w->core.width, &w->core.height ) ;

  /* ɥΥꥵ¾ꡣ*/
  sizehints.win_gravity = NorthWestGravity ;
  sizehints.base_width  = canvas_popup->core.border_width * 2 ;
  sizehints.base_height = canvas_popup->core.border_width * 2 ;
  sizehints.width      = winwidth ;
  sizehints.height     = winheight ;
  sizehints.width_inc  = w->separateWin.roman_width ;
  sizehints.height_inc = w->separateWin.font_height ;
  sizehints.min_width  = w->separateWin.roman_width * 2 +
    sizehints.base_width ;
  sizehints.min_height = w->separateWin.font_height * 3 +
    sizehints.base_height ;
  sizehints.flags      = 
    ( PBaseSize | PMinSize | PResizeInc | PMaxSize | USSize | PWinGravity ) ;

  /* Window Manager ˥ҥȤäƤ롣*/
  XSetWMNormalHints
    ( XtDisplay( gw ), XtWindow( canvas_popup ), &sizehints ) ;

  if( w->separateWin.historyAttribute != NULL ){
    SEPW_CanvasPopupPositionInit
      ( gw, w->separateWin.historyAttribute->client,
	winwidth, winheight, canvas_popup->core.border_width ) ;
  }

  XtPopup( canvas_popup, XtGrabNone ) ;
  w->separateWin.canvas_probe = True ;
  XtSetKeyboardFocus( canvas_popup, gw ) ;

  /* ̤κ*/
  XtVaSetValues
    ( canvas, XtNredrawRequest, True, NULL ) ;
#if 0
  SEPW_RedrawScreen( gw ) ;
#endif
  XSync( display, False ) ;
  return ;
}

/*
 * SIW  XtDestroyWidget 򤫤˸ƤӽФؿ
 *-----
 * SIW νλ·äƤ롣GC Ƥ顢malloc Ƥ
 * Ƥ롣
 */
static void SEPW_Destroy( Widget gw )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  struct SKKInputNode *node, *pNode ;
  /*
   * XFreeFont( XtDisplay( gw ), w->separateWin.fs_kanji ) ;
   * XFreeFont( XtDisplay( gw ), w->separateWin.fs_roman ) ;
   * ɤ顢Widget ˳ݤƤ櫓ʤߤʤΤǡ
   * ȤϤޤߤ
   */
  ignore_xerror() ;
  /* Event ʤ֤ˤƤƤġ*/
  XSelectInput( XtDisplay( gw ), XtWindow( gw ), NoEventMask ) ;
  /* XtDestroyWidget( w->separateWin.canvas_popup ) ;
     XtDestroyWidget( w->separateWin.canvas ) ; */
  /* Event Queue եå夷Ƥ*/
  XSync( XtDisplay( gw ), True ) ;
  regard_xerror() ;

  XtReleaseGC( gw, w->separateWin.roman_ngc ) ;
  XtReleaseGC( gw, w->separateWin.roman_rgc ) ;
  XtReleaseGC( gw, w->separateWin.kanji_ngc ) ;
  XtReleaseGC( gw, w->separateWin.kanji_rgc ) ;

  node = w->separateWin.lastbuffer ;
  while( node != NULL ){
    pNode = node->parentbuffer ;
    free_Minibuffer
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, node ) ;
    node  = pNode ;
  }
  if( w->separateWin.historyAttribute != NULL ){
    w->separateWin.historyAttribute->history_start =
      w->separateWin.hist_start ;
    w->separateWin.historyAttribute->history_end =
      w->separateWin.hist_end ;
  }
  XtCallCallbacks( gw, XtNendNotify, w->separateWin.historyAttribute ) ;
  w->separateWin.historyAttribute = NULL ;
  return ;
}

/*
 * ̤褹ؿ
 */
static void SEPW_Redisplay( Widget gw, XEvent *event, Region region )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  /* ϲΥߥ󥰤Ǵְä Widget κ褬Ƥ롣*/
  if( w->separateWin.topbuffer == NULL || w->core.being_destroyed ){
#if 1
    fprintf( stderr, "Illegal Widget Exposure Event.\n" ) ;
#endif
    return ;
  }
  XtVaSetValues
    ( w->separateWin.canvas, XtNredrawRequest, True, NULL ) ;
#if 0
  SEPW_RedrawScreen( gw ) ;
#endif
/*  XFlush( XtDisplay( gw ) ) ;*/
  return ;
}

static Boolean SEPW_SetValues
( Widget current, Widget request, Widget new,
  ArgList args, Cardinal *num_args )
{
  SeparateWinWidget curw = ( SeparateWinWidget )current ;
  SeparateWinWidget reqw = ( SeparateWinWidget )request ;
  SeparateWinWidget neww = ( SeparateWinWidget )new ;

  if( reqw->separateWin.historyAttribute != NULL ){
    /* ҥȥν򤹤롣ɬ¹Ԥ롣*/
    neww->separateWin.historybuffer =
      reqw->separateWin.historyAttribute->history ;
    neww->separateWin.hist_start = 
      reqw->separateWin.historyAttribute->history_start ;
    neww->separateWin.hist_end = 
      reqw->separateWin.historyAttribute->history_end ;
    return FALSE ;
  }

  /* ХƤʤСʲν̵Ǥ롣*/
  if( !neww->separateWin.canvas_probe )
    return FALSE ;

  /* 񤬽Ƥ顢ľ¹Ԥ롣*/
  if( curw->separateWin.jisyo_dirty != reqw->separateWin.jisyo_dirty ){
    neww->separateWin.jisyo_dirty = reqw->separateWin.jisyo_dirty ;
    XtVaSetValues
      ( neww->separateWin.canvas, XtNredrawRequest, True, NULL ) ;
    return FALSE ;
  }
  if( reqw->separateWin.clearMinibuffer ){
    SEPW_ClearMinibuffer( new ) ;
    return ( TRUE ) ;
  }
  return( FALSE ) ;
}

enum {
  SIW_DESTROYED = -1, SIW_PROCESSING = 0,
} ;

/*
 * 줿Ȥ٥ȤνԤؿ
 *----
 * Υ줿ȤԲǽʾ礬Τǡ줿Ȥ
 * Ǿ꤯ư褦˺ʤФʤʤ
 */
static void KeyDownEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  char inpstr[ STRBUFSIZE ] ;
  KeySym key ;
  struct SKKInputNode *node ;

  /* 줿ΤǤʤȴ롣*/
  if( xevent->type != KeyPress )
    return ;

  /* ٥Ȥ򵭲Ƥ*/
  w->separateWin.xevent = *xevent ;

  /* ߥ˥Хåե¸ߤƤΤʤСΰֺǸΤΤ򤹤롣*/
  node = w->separateWin.lastbuffer ;
  if( !node->cur_exist )
    node = w->separateWin.topbuffer ;

  /* ʸХåեν*/
  inpstr[ 0 ] = '\0' ;
  /* ʸ*/
  XLookupString( &( xevent->xkey ), inpstr, STRBUFSIZE, &key, NULL ) ;
  
  /* եȥڡξνԤ*/
  if( ( xevent->xkey.state & ShiftMask ) && key == XK_space &&
     w->separateWin.compatible_close_skkinputkey ){
    XtDestroyWidget( gw ) ;
    return ;
  }

  /* Ϥ줿ʸϰ̣ʸʤΤǤ */
  if( inpstr[ 0 ] != '\0' ){
    w->separateWin.keybuf[ w->separateWin.keybuf_use ++ ] = inpstr[ 0 ] ;
    /* ߥ˥Хåե¸ߤѤƤν*/
    switch( do_function_by_reading_character( gw, node ) ){
    case SIW_DESTROYED :
      XtDestroyWidget( gw ) ;
      return ;
    case SIW_PROCESSING :
      break ;
    default :
      return ;
    }
  } else {
    /* ȥ륹ڡξνԤ string  */
    /* ʤȦʤΤǡ*/
    if( ( xevent->xkey.state & ControlMask ) && key == XK_space ){
      w->separateWin.keybuf[ w->separateWin.keybuf_use ++ ] = 0x00 ;
      if( do_function_by_reading_character( gw, node ) )
	return ;
    }
  }
  skkinput_DrawScreen( gw ) ;
/*  XFlush( XtDisplay( gw ) ) ;*/
  return ;
}

/*
 * ե줿νԤؿ
 */
static void FocusOutEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  /* ե줿Ȥã줿Τǡ򵭲Ƥ*/
  w->separateWin.is_focus = False ;
  /* ̤νĥηѲΤǡ*/
  skkinput_DrawScreen( gw ) ;
  return ;
}

/*
 * ե蘆줿νԤؿ
 */
static void FocusInEventHandler
( Widget gw, XEvent *xevent, String *params, Cardinal *num_params )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  /* ե줿Ȥã줿Τǡ򵭲Ƥ*/
  w->separateWin.is_focus = True ;
  /* ̤νĥηѲΤǡ*/
  skkinput_DrawScreen( gw ) ;
  return ;
}

/*
 * ʲΥեϡSeparateWinWidget ΥΰǤ롣
 */

/*
 * ơ֥򻲾Ȥ̤Υե󥯥ֹ˽äƽ¹Ԥ
 * ٤¾δؿƽФԤؿ
 */
static int do_skkinputFunction
( Widget gw, struct SKKInputNode *node,	int func_no )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int chara = w->separateWin.keybuf[ 0 ] ;

  /* ȡΥåХåե˽񤫤Ƥ *
   * Ͼä뤳ȤˤʤäƤ롩 */
  node->mtextbuffer[ 0 ] = '\0' ;

  /* 䴰ײλ뤫ɤ*/
  if( node->j_completion_mode ){
    if( func_no != FUNCNO_J_INSERT_PERIOD &&
	func_no != FUNCNO_J_INSERT_COMMA  &&
	func_no != FUNCNO_J_ABBREV_PERIOD &&
	func_no != FUNCNO_J_ABBREV_COMMA  ){
      j_completion_close( node ) ;
    }
  }
  /* Ϳ줿̿ФԤ*/
  switch( func_no ){
  case FUNCNO_J_PREFIX_CHAR :
  case FUNCNO_J_SENDBACK_KEY :
    j_delete_kanaprefix( node ) ;
    /* ֢ΥХåեʤСC- Ȥɽ롣*/
    if( node == ( w->separateWin ).topbuffer ){
      strcpy( node->mtextbuffer,
	     charaToString( w->separateWin.keybuf[ 0 ] ) ) ;
      strcat( node->mtextbuffer, "-" ) ;
    }
    return SIW_PROCESSING ;
  case FUNCNO_SELF_INSERT_CHARACTER :
    j_insert_chara( node, node->cur_pos, chara ) ;
    break ;
  case FUNCNO_J_SELF_INSERT :
    j_self_insert
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_SELF_ZENKAKU_INSERT :
    j_insert_zenkaku( node, node->cur_pos, chara ) ;
    break ;
  case FUNCNO_J_DISPLAY_CODE_FOR_CHAR_AT_POINT :
    j_display_code_for_char_at_point( node ) ;
    break ;
  case FUNCNO_J_SET_HENKAN_POINT :
    /* Ѵϰ֤ꤵ롣*/
    j_set_henkan_point
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_SET_HENKAN_POINT_SUBR :
    j_set_henkan_point_subr
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, 
	node, chara ) ;
    break ;
    /* 첻ϡ*/
  case FUNCNO_J_INSERT_A :
    j_insert_a
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_E :
    j_insert_e
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_I :
    j_insert_i
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_O :
    j_insert_o
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_INSERT_U :
    j_insert_u
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
    /* ̾Ϥγϥ( ҲǤ )*/
  case FUNCNO_J_KANA_INPUT :
    j_kana_input
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_START_HENKAN :
    j_start_henkan
      ( gw, &w->separateWin.topbuffer,
        &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INSERT_COMMA :
    j_insert_comma
      ( gw, &w->separateWin.topbuffer,
       &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INSERT_PERIOD :
    j_insert_period
      ( gw, &w->separateWin.topbuffer,
       &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_PURGE_FROM_JISYO :
    j_purge_from_jisyo( gw, &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_INPUT_BY_CODE_OR_MENU :
    j_input_by_code_or_menu_start
      ( gw, &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_MODE_OFF :
    j_mode_off
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, chara ) ;
    break ;
  case FUNCNO_J_MODE_OFF_AND_SELF_INSERT :
    j_mode_off_and_self_insert
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
 	node, chara ) ;
    break ;
  case FUNCNO_J_TOGGLE_KANA :
    /* ̾⡼ɤҲ̾⡼ɤڤؤ*/
    j_toggle_kana
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, 
	node, False, chara ) ;
    break ;
  case FUNCNO_J_PREVIOUS_CANDIDATE :
    j_previous_candidate
      ( gw, &w->separateWin.topbuffer,
       &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_KAKUTEI :
    if( j_kakutei_kana_mode
	( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, 
	  node, chara ) )
      break ;
    j_kakutei( gw, node ) ;
    break ;
  case FUNCNO_J_ABBREV_INPUT :
    j_abbrev_input
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_ABBREV_PERIOD :
    j_abbrev_period( gw, node ) ;
    break ;
  case FUNCNO_J_ABBREV_COMMA :
    j_abbrev_comma( gw, node ) ;
    break ;
  case FUNCNO_J_ZENKAKU_EIJI :
    /* (Ѥܸ)ѻϥ⡼ɤˤ롣*/
    j_zenkaku_eiji
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, 
	node, chara ) ;
    break ;
  case FUNCNO_J_ZENKAKU_HENKAN :
    j_zenkaku_henkan( gw, node ) ;
    break ;
  case FUNCNO_J_TODAY :
    /* դνϡ*/
    j_today( node ) ;
    break ;
  case FUNCNO_J_KANAINPUT_MODE_ON :
    j_kanainput_mode_on
      ( gw, &w->separateWin.topbuffer,
       &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_J_SET_MARK_COMMAND :
    j_set_mark_command( node ) ;
    break ;
  case FUNCNO_J_NEWLINE :
    j_newline
      ( gw, &( w->separateWin.xevent ), False,
	&w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, w->separateWin.egg_like_newline,
	w->separateWin.chat_adapter,
	w->separateWin.historybuffer, 
	&w->separateWin.hist_start, &w->separateWin.hist_end,
	&w->separateWin.hist_cur ) ;
    break ;
  case FUNCNO_J_FORWARD_CHAR :
    j_forward_char( node ) ;
    break ;
  case FUNCNO_J_BACKWARD_CHAR :
    j_backward_char( node ) ;
    break ;
  case FUNCNO_J_DELETE_CHAR :
    j_delete_char( node ) ;
    break ;
  case FUNCNO_J_DELETE_BACKWARD_CHAR :
    j_backward_delete_char
      ( gw, &w->separateWin.topbuffer,
        &w->separateWin.lastbuffer, node,
        &w->separateWin.xevent, True ) ;
    break ;
  case FUNCNO_J_TRY_COMPLETION :
    j_try_completion( gw, node ) ;
    break ;
  case FUNCNO_J_END_OF_LINE :
    j_end_of_line( node ) ;
    break ;
  case FUNCNO_J_BEGINNING_OF_LINE :
    j_beginning_of_line( node ) ;
    break ;
  case FUNCNO_J_KILL_LINE :
    j_kill_line( node, w->separateWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_KILL_REGION :
    j_kill_region( node, w->separateWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_KILL_RING_SAVE :
    j_kill_ring_save( node, w->separateWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_EXCHANGE_POINT_AND_MARK :
    j_exchange_point_and_mark( node ) ;
    break ;
  case FUNCNO_J_YANK :
    j_yank( node, w->separateWin.cutbuffer ) ;
    break ;
  case FUNCNO_J_PREVIOUS_LINE :
    j_previous_line
      ( gw, 
	&w->separateWin.topbuffer, &w->separateWin.lastbuffer, node,
	w->separateWin.historybuffer, w->separateWin.histcurbackbuffer, 
	w->separateWin.hist_start, w->separateWin.hist_end,
	&w->separateWin.hist_cur ) ;
    break ;
  case FUNCNO_J_NEXT_LINE :
    j_next_line
      ( gw, 
	&w->separateWin.topbuffer, &w->separateWin.lastbuffer, node,
	w->separateWin.historybuffer, w->separateWin.histcurbackbuffer, 
	w->separateWin.hist_start, w->separateWin.hist_end,
	&w->separateWin.hist_cur ) ;
    break ;
  case FUNCNO_J_TRANSPOSE_CHARS :
    j_transpose_chars( node ) ;
    break ;
  case FUNCNO_J_REDRAW :
    /* öե褹롣ʬǤϤʤơ*/
    XtVaSetValues
      ( w->separateWin.canvas, XtNredrawRequest, True, NULL ) ;
#if 0
    SEPW_RedrawScreen( gw ) ;
#endif
    break ;
  case FUNCNO_J_KEYBOARD_QUIT :
    j_keyboard_quit
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer, node ) ;
    break ;
  case FUNCNO_SAVE_SKKINPUT_LOCAL_JISYO :
    j_delete_kanaprefix( node ) ;
    /* 񤬹Ƥ뤫ɤȽꤹ롣*/
    if( w->separateWin.jisyo_dirty ){
      strcpy( node->mtextbuffer, "Saving Jisyo..." ) ;
      skkinput_DrawScreen( gw ) ;
      /* XFlush( XtDisplay( gw ) ) ;*/
      /* ι򤹤륳ХåƽФ*/
      skkinput_updateLocalJisyo() ;
      
      strcat( node->mtextbuffer, "done" ) ;
    } else {
      err_msg( node, "(No changes need to be saved)" ) ;
    }
    break ;
  case FUNCNO_CLOSE_SKKINPUT :
    return SIW_DESTROYED ;
  case FUNCNO_VC_TOGGLE_CHATMODE :
    j_delete_kanaprefix( node ) ;
    w->separateWin.chat_adapter =
      ( w->separateWin.chat_adapter )? False : True ;
    break ;
  case FUNCNO_VC_TOGGLE_EGGNL :
    j_delete_kanaprefix( node ) ;
    w->separateWin.egg_like_newline =
      ( w->separateWin.egg_like_newline )? False : True ;
    break ;
  case FUNCNO_NOT_MODIFIED :
    j_delete_kanaprefix( node ) ;
    /* ѹե饰õ롣*/
    j_clear_modified_flag() ;
    strcpy( node->mtextbuffer, "Modification-flag cleared" ) ;
    break ;
  case FUNCNO_J_FORWARD_WORD :
    j_forward_word( node ) ;
    break ;
  case FUNCNO_J_BACKWARD_WORD :
    j_backward_word( node ) ;
    break ;
  case FUNCNO_J_UPCASE_WORD :
    j_caseword( node, 'u' ) ;
    break ;
  case FUNCNO_J_DOWNCASE_WORD :
    j_caseword( node, 'd' ) ;
    break ;
  case FUNCNO_J_CAPITALIZE_WORD :
    j_caseword( node, 'c' ) ;
    break ;
  case FUNCNO_INVALID_CHAR :
  default :
#if 0
    XBell( XtDisplay( gw ), 5 ) ;
#endif
    break ;
  }
  w->separateWin.keybuf_use  = 0 ;
  w->separateWin.keybuf[ 0 ] = '\0' ;

  return SIW_PROCESSING ;
}

/*
 * ĤξνԤؿ
 */
static int do_function_by_onekey
( Widget gw, struct SKKInputNode *node )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int func_no ;

  /* Ѵθ⡼ɤʤΤǤ */
  if( node->j_henkan_show_candidate_mode ){
    /* ɤʥäϤͤȤäɡĤޤΥ⡼*
     * ɤФäƤ뤫פʡ */
    do_Function_jhenkanShowCandidateMode
      ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	node, w->separateWin.keybuf ) ;
    /* ʸΥХåեõ롣*/
    w->separateWin.keybuf_use  = 0 ;
    w->separateWin.keybuf[ 0 ] = '\0' ;
    return SIW_PROCESSING ;
  } else if( node->j_input_by_code_or_menu_mode ){
    if( node->j_input_by_code_or_menu_mode == True ){
      do_function_j_input_by_code_or_menu_jump
	( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	  node, w->separateWin.keybuf ) ;
      /* ʸΥХåեõ롣*/
      w->separateWin.keybuf_use  = 0 ;
      w->separateWin.keybuf[ 0 ] = '\0' ;
      return SIW_PROCESSING ; 
    } else {
      do_function_j_input_by_code_or_menu_1_jump
	( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	  node, w->separateWin.keybuf ) ;
      /* ʸΥХåեõ롣*/
      w->separateWin.keybuf_use  = 0 ;
      w->separateWin.keybuf[ 0 ] = '\0' ;
      return SIW_PROCESSING ; 
    }
  }
  /* Ϥ줿ǽʸ򸫤ơĵǽʤǤäФ*/
  func_no = lookup_at_keymap( w->separateWin.keybuf, node ) ;

  /* 
   * ¿ʬʳ skk-rom-kana-rule-list åʤȤ
   * ǤʤХȥ륳ɤǤ⤽줬ȤȤȤ
   * Ϥʤʤ顣
   */
  switch( func_no ){
  case FUNCNO_J_INSERT_A :
  case FUNCNO_J_INSERT_E :
  case FUNCNO_J_INSERT_I :
  case FUNCNO_J_INSERT_O :
  case FUNCNO_J_INSERT_U :
  case FUNCNO_J_KANA_INPUT :
  case FUNCNO_J_SET_HENKAN_POINT :
    break ;
  default :
    /* ̾Ϥäꤷˤϡ⤷   *
     * rom-kana-rule-list ˥ҥåȤꤹ뤫⤷ʤ*/
    if( node->j_kana_mode ){
      if( rom_kana_rule_list_check
	 ( node, w->separateWin.keybuf[ 0 ] ) ){
	if( node->j_henkan_mode && !node->j_henkan_on &&
	   node->j_okurigana_mode ){
	  j_set_okurigana
	    ( gw, &w->separateWin.topbuffer, &w->separateWin.lastbuffer,
	      node, node->j_okuri_chara ) ;
	}
	w->separateWin.keybuf_use  = 0 ;
	w->separateWin.keybuf[ 0 ] = '\0' ;
	return SIW_PROCESSING ;
      }
    }
    /*  oh Υå롣*/
    j_oh_check_with_controlcode
      ( node, ( ( func_no == FUNCNO_J_KEYBOARD_QUIT )? True : False ) ) ;
    break ;
  }
  /* Ϳ줿ǽ¹פޤ*/
  return do_skkinputFunction( gw, node, func_no ) ;
}

/*
 * ĤξνԤؿ
 */
static int do_function_by_twokeys
( Widget gw, struct SKKInputNode *node )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;

  /* ȡΥåХåե˽񤫤Ƥ *
   * Ͼä뤳ȤˤʤäƤ롩 */
  node->mtextbuffer[ 0 ] = '\0' ;

  /* Ϥ줿Ϥ⤷ơ */
  if( lookup_at_keymap( w->separateWin.keybuf, node ) ==
      FUNCNO_J_SENDBACK_KEY ){
    j_sendback_keypress( gw, &w->separateWin.xevent ) ;
  } else {
    /* ơ֥ܤʣξϤʤ *
     * õʤƴŤȤޤ*/
    return do_skkinputFunction
      ( gw, node, lookup_at_twokeymap( w->separateWin.keybuf ) ) ;
  }
  /* ʾʸФʤΤǡľ*/
  w->separateWin.keybuf[ 0 ] = '\0' ;
  w->separateWin.keybuf_use  = 0 ;
  return SIW_PROCESSING ;
}

/*
 * Ϥ줿ʸ󤫤餽б뵡ǽ¹Ԥؿ
 */
static int do_function_by_reading_character
( Widget gw, struct SKKInputNode *node )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;

#ifdef DEBUG
  fprintf( stderr, "j-prefix : %s\n", node->j_prefix ) ;
#endif

  switch( w->separateWin.keybuf_use ){
    /* ХϤФ*/
  case 1:
    return do_function_by_onekey( gw, node ) ;
    /* ʣХȤΥϤФ*/
  case 2:
    return do_function_by_twokeys( gw, node ) ;
    /* ꤨʤϤФ*/
  case 0 :
  default :
    w->separateWin.keybuf_use  = 0 ;
    w->separateWin.keybuf[ 0 ] = '\0' ;
    break ;
  }
  return SIW_PROCESSING ;
}

enum {
  MODELINE_NOTHING = 0, MODELINE_HANKAKU, MODELINE_ZENKAKU,
  MODELINE_KATAKANA, MODELINE_KANA,
} ;

static void skkinput_FullDrawWindow
( Widget gw, struct SKKInputNode *node, char *text, 
  int start_pos, int x, int y, int end_x, int end_y ) ;

static char *skkinput_WinGetNextLineTop
( char *ptr, int win_width, int roman_width, int kana_width )
{
  int kana_flag = 0, x = 0 ;

  while( *ptr != '\0' ){
    if( 0x80 & *ptr ){
      kana_flag = kana_flag % 2 + 1 ;
    } else {
      kana_flag = 0 ;
    }
    switch( kana_flag ){
      /* ɤν*/
    case 0 :
      x += roman_width ;
      if( x >= win_width )
	return ptr ;
      if( *ptr == '\n' )
	return ( ptr + 1 ) ;
      break ;
      /* ɤΰХܤν*/
    case 1 :
      x += kana_width ;
      if( x >= win_width )
	return ptr ;
      break ;
      /* ɤХܤν*/
    case 2 :
    default :
      break ;
    }
    ptr ++ ;
  }
  return ptr ;
}

/*
 * ʬԤäƤʬΥᥤʬĤδؿ
 *---------------
 * ɽ줿ʸȺɽʤȤʤʸ򤸤äȸ٤
 * ɤʡ ʸƱɽʤǺѤ͡ ȤԤ
 * ä顢ѹϤ͡ Ȥġ
 */
static char *skkinput_DiffDrawWindowSub1
( Widget gw, struct SKKInputNode *node, char *text, char *ptext,
  int start_pos, int *retx, int *rety, int end_x, int end_y,
  int *rcx, int *rcy, int *retpos )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int x, y, pos, cx, cy, chara ;
  int charas_are_different, dkana_flag, skana_flag, diff_flag ;
  char *dptr, *sptr, *linetop ;

  /* ե饰ν*/
  diff_flag = skana_flag = dkana_flag = False ;

  /* ɽ֤ꡣ*/
  pos = start_pos ;
  x = *retx ;
  y = *rety ;
  chara = 0 ;
  cx = cy = -1 ;

  /* ʸꡣ*/
  linetop = sptr = ptext ;
  dptr = text ;

  /* ĤʸӤ롣*/
  while( *dptr != '\0' ){
    /* ɤɤɤʤ餽βХܤ*/
    if( 0x80 & *dptr ){
      dkana_flag = dkana_flag % 2 + 1 ;
    } else {
      dkana_flag = 0 ;
    }
    /* Ԥ뤫ɤȽꡣ*/
    switch( dkana_flag ){
      /* ɤν*/
    case 0 :
      /* Ԥޤ֤Υå*/
      if( ( x + w->separateWin.roman_width ) >= end_x ){
	/* 뤬ɽƤ顢äѤõ*/
	if( x == node->prev_cx && y == node->prev_cy )
	  skkwin_jerase
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_rgc, x, y, 
	      w->separateWin.font_ascent,
	      w->separateWin.kanji_width,
	      w->separateWin.font_height ) ;
	/* ޤ֤򼨤 backslash ɽ롣*/
	skkwin_jputchar
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	    x, y, '\\', False ) ;
	x = 0 ;
	y += w->separateWin.font_height ;
	linetop = sptr = skkinput_WinGetNextLineTop
	  ( linetop, end_x, w->separateWin.roman_width,
	    w->separateWin.kanji_width ) ;
	/* ǲԤ뤳ȤϤʤΤǡǤ롣*/
	skana_flag = 0 ;
	/* ιԿĶƤޤäʤ顢ȴ롣*/
	if( y >= end_y )
	  goto exit_loop ;
      }
      break ;
      /* ɤΰХܤν*/
    case 1 :
      if( ( x + w->separateWin.kanji_width ) >= end_x ){
	/* 뤬ɽƤ顢äѤõ*/
	if( x >= node->prev_cx &&
	    ( node->prev_cx <= ( x + w->separateWin.kanji_width ) ) &&
	    y == node->prev_cy )
	  skkwin_jerase
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_rgc, x, y, 
	      w->separateWin.font_ascent,
	      w->separateWin.kanji_width,
	      w->separateWin.font_height ) ;
	skkwin_jputchar
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	    w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	    x, y, '\\', False ) ;
	skkwin_jputchar
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	    x + w->separateWin.roman_width, y, '\\', False ) ;
#if 0
	skkwin_jerase
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_rgc, x, y, 
	    w->separateWin.font_ascent,
	    w->separateWin.kanji_width,
	    w->separateWin.font_height ) ;
#endif
	x = 0 ;
	y += w->separateWin.font_height ;
	/* ӤƤɽ줿ʸԤ줿֤õơ*
	 * إݥ󥿤ư롣*/
	linetop = sptr = skkinput_WinGetNextLineTop
	  ( linetop, end_x, w->separateWin.roman_width,
	    w->separateWin.kanji_width ) ;
	/* ǲԤ뤳ȤϤʤΤǡǤ롣*/
	skana_flag = 0 ;
	/* ιԿĶƤޤäʤ顢ȴ롣*/
	if( y >= end_y )
	  goto exit_loop ;
      }
      break ;
      /* ɤХܤν*/
    case 2 :
    default :
      break ;
    }
    /* ɽ줿ȽԤ*/
    if( 0x80 & *sptr ){
      skana_flag = skana_flag % 2 + 1 ;
    } else {
      skana_flag = 0 ;
    }
    /* ơ饯ϰۤʤΤǤ礦 */
    if( *sptr != *dptr || pos == node->cur_pos ||
	( x == node->prev_cx && y == node->prev_cy ) ||
	skana_flag != dkana_flag ){
      /* ࡣ֤ϥǥեȤǰۤʤΤʡ */
      charas_are_different = True ;
    } else {
      charas_are_different = False ;
    }
    switch( dkana_flag ){
      /* ɤν*/
    case 0 :
      /* ꤬˲ԤƤޤäƤˤ̾ɽ*/
      /* ǤϤʤơ饯Ӥưۤʤɽ*/
      if( charas_are_different ){
	if( pos == node->cur_pos ){
	  if( node->cur_exist ){
	    skkinputDrawCursor
	      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
		w->separateWin.roman_ngc, w->separateWin.roman_rgc,
		x, y, w->separateWin.font_ascent,
		w->separateWin.roman_width,
		w->separateWin.font_height,
		w->separateWin.is_focus ) ;
	    if( w->separateWin.is_focus ){
	      skkwin_jputchar
		( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
		  w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
		  x, y, *dptr, True ) ;
	    } else {
	      skkwin_jputchar
		( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
		  w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
		  x, y, *dptr, True ) ;
	    }
	  } else {
	    skkwin_jputchar2
	      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
		w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
		w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
		x, y, *dptr,
		w->separateWin.roman_width,
		w->separateWin.kanji_width,
		w->separateWin.font_height,
		w->separateWin.font_ascent, False ) ;
	  }
	  cx = x ; cy = y ;
	} else {
	  skkwin_jputchar2
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	      w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	      x, y, *dptr,
	      w->separateWin.roman_width,
	      w->separateWin.kanji_width,
	      w->separateWin.font_height,
	      w->separateWin.font_ascent, False ) ;
	}
      }
      /* ɽ֤ΰư*/
      x += w->separateWin.roman_width ;
      if( *dptr == '\n' ){
	/* ϴ˲ԤƤΤǤ礦 */
	if( *sptr != '\n' && *sptr != '\0' ){
	  skkwin_jerase
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_rgc, x, y, 
	      w->separateWin.font_ascent,
	      end_x - x,
	      w->separateWin.font_height ) ;
	}
	/* Ƭ˰ư롣*/
	x = 0 ;
	y += w->separateWin.font_height ;
	linetop = sptr = skkinput_WinGetNextLineTop
	  ( linetop, end_x, w->separateWin.roman_width,
	    w->separateWin.kanji_width ) ;
	/* ǲԤ뤳ȤϤʤΤǡǤ롣*/
	skana_flag = 0 ;
	/* ιԿĶƤޤäʤ顢ȴ롣*/
	if( y >= end_y )
	  goto exit_loop ;
      }
      break ;
      /* ɤΰХܤν*/
    case 1 :
      chara = ( unsigned char )*dptr ;
      /* ä˲ԤƤޤäƤˤԤġ*/
      /* ɤΰХܤʸ󤬰ۤʤäƤ硣*/
      if( charas_are_different )
	diff_flag = True ;
      break ;
      /* ɤХܤν*/
    case 2 :
      chara = ( chara << 8 ) | ( ( unsigned char )*dptr ) ;
      /* ɤХܤʸ󤬰ۤʤäƤڤӡХ *
       * ܤǰۤʤäƤˤϽľκȤ롣                 */
      if( charas_are_different || diff_flag ){
	if( pos == ( node->cur_pos + 1 ) ){
	  if( node->cur_exist ){
	    skkinputDrawCursor
	      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
		w->separateWin.roman_ngc, w->separateWin.roman_rgc,
		x, y, w->separateWin.font_ascent,
		w->separateWin.kanji_width,
		w->separateWin.font_height,
		w->separateWin.is_focus ) ;
	    if( w->separateWin.is_focus ){
	      skkwin_jputchar
		( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
		  w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
		  x, y, chara, True ) ;
	    } else {
	      skkwin_jputchar
		( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
		  w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
		  x, y, chara, True ) ;
	    }
	  } else {
	    skkwin_jputchar2
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	      w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	      x, y, chara,
	      w->separateWin.roman_width,
	      w->separateWin.kanji_width,
	      w->separateWin.font_height,
	      w->separateWin.font_ascent, False ) ;
	  }
	  cx = x ; cy = y ;
	} else {
	  skkwin_jputchar2
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	      w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	      x, y, chara,
	      w->separateWin.roman_width,
	      w->separateWin.kanji_width,
	      w->separateWin.font_height,
	      w->separateWin.font_ascent, False ) ;
	}
	diff_flag = False ;
      }
      /* ɽ֤ΰư*/
      x += w->separateWin.kanji_width ;
      break ;
    default :
      dkana_flag = 0 ;
      break ;
    }
    dptr ++ ;
    if( *sptr != '\0' && *sptr != '\n' )
      sptr ++ ;
    pos  ++ ;
  }
exit_loop:
  /* ɽη̥뤬ޤǰưΤ֤*/
  *retx   = x ;
  *rety   = y ;
  *rcx    = cx ;
  *rcy    = cy ;
  *retpos = pos ;
  return sptr ;
}

/*
 * ʬɽԤؿ
 *----
 * ޤޤưʤǤɤƤϡ档
 */
static void skkinput_DiffDrawWindow
( Widget gw, struct SKKInputNode *node, char *text, char *ptext,
  int start_pos, int x, int y, int end_x, int end_y )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int pos, cx, cy, kx, ky ;
  char *sptr ;

  /* ξԤɽƤʬˤĤƺʬԤ*/
  sptr = skkinput_DiffDrawWindowSub1
    ( gw, node, text, ptext, start_pos, &x, &y, end_x, end_y,
      &cx, &cy, &pos ) ;

  /* ޤ̰դˤϽ񤤤ƤʤΤǡõȤĤäƤ뤫⤷ʤ*/
  if( y < end_y ){
    /* õԤ˸ߤΥΰ֤ФƤ*/
    kx = x ;
    ky = y ;
    while( *sptr != '\0' ){
      if( ( ( x + w->separateWin.kanji_width ) >= end_x &&
	    ( 0x80 & *sptr ) ) ||
	  ( x + w->separateWin.roman_width ) >= end_x ){
	skkwin_jerase
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_rgc, x, y, 
	    w->separateWin.font_ascent,
	    w->separateWin.kanji_width,
	    w->separateWin.font_height ) ;
	x = 0 ;
	y += w->separateWin.font_height ;
      }
      if( y >= end_y )
	break ;
      skkwin_jerase
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	  w->separateWin.roman_rgc, x, y, 
	  w->separateWin.font_ascent,
	  w->separateWin.roman_width,
	  w->separateWin.font_height ) ;
      x += w->separateWin.roman_width ;
      sptr ++ ;
    }

    /* 뤬ɽƤ֤ޤãƤޤ... *
     * ˤϡõޤ*/
    if( y < node->prev_cy ||
	( x <= node->prev_cx && y == node->prev_cy ) ){
      skkwin_jerase
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	  w->separateWin.roman_rgc, node->prev_cx, node->prev_cy, 
	  w->separateWin.font_ascent,
	  w->separateWin.kanji_width,
	  w->separateWin.font_height ) ;
    }
    x = kx ;
    y = ky ;

    /* ޤ뤬ɽƤʤȤȤϡǸΰ֤ˤ *
     * Τʤ*/
    if( cy < 0 ){
      if( ( x + w->separateWin.roman_width ) > end_x ){
	x = 0 ;
	y += w->separateWin.font_height ;
      }
      if( node->cur_exist )
	skkinputDrawCursor
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_ngc, w->separateWin.roman_rgc,
	    x, y, w->separateWin.font_ascent,
	    w->separateWin.roman_width,
	    w->separateWin.font_height,
	    w->separateWin.is_focus ) ;
      cx = x ; cy = y ;
    }
  }
  /* 󥫡ɽ֤򵭲Ƥ*/
  node->prev_cx = cx ; 
  node->prev_cy = cy ;
  /* ɽʸ򵭲Ƥ*/
  return ;
}

/*
 * Ϥ줿 skkinput Υ⡼ɥ饤󤬲ǤΤĴ٤ؿ
 */
static int skkinput_GetModelineNumber( struct SKKInputNode *node )
{
  int modeline ;
  /* ̾ϥ⡼ɤʤΡ ǤϤʤΡ */
  if( !node->j_mode ){
    /* ϥ⡼ɤʤΡ Ⱦϥ⡼ɤʤΡ */
    if( node->j_zenkaku ){
      modeline = MODELINE_ZENKAKU ;
    } else {
      modeline = MODELINE_HANKAKU ;
    }
  } else {
    /* Ҳ̾ϥ⡼ɤʤΡ ʿ̾ϥ⡼ɤʤΡ */
    if( node->j_katakana_mode ){
      modeline = MODELINE_KATAKANA ;
    } else {
      modeline = MODELINE_KANA ;
    }
  }
  return modeline ;
}

/*
 * skkinput Υ⡼ɥ饤ʬ褹ؿ
 */
static void skkinput_DiffDrawModeLine
( Widget gw, struct SKKInputNode *node, int x, int y, unsigned int width )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  static char *modelineStrings[] = {
    "SKK:", "", "", "",
  } ;
  static char *modelineStrings2[] = {
    "--", "**",
  } ;
  static char *chatAdapterStrings[] = {
    "(Fundamental",
    "(ChatAdapter",
  } ;
  int modeline = skkinput_GetModelineNumber( node ) ;

  /* ⡼ɥ饤ɽ줿ΤƱǤ뤫ɬפ *
   * ʤ*/
  if( w->separateWin.prev_modeline != modeline ){
    /* ѹоݤȤʤʬõ롣*/
    XFillRectangle
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	w->separateWin.roman_ngc,
	x + w->separateWin.roman_width * 2,
	y - w->separateWin.font_ascent,
	w->separateWin.kanji_width * 2,
	w->separateWin.font_height ) ;
    /* ⡼ɥ饤ʬ褹롣*/
    skkwin_jputstring
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
	x + w->separateWin.roman_width * 2, y,
	modelineStrings[ modeline - 1 ],
	w->separateWin.roman_width,
	w->separateWin.kanji_width ) ;
    /* ɽФ롣*/
    w->separateWin.prev_modeline = modeline ;
  }
  if( w->separateWin.prev_jisyo_dirty != w->separateWin.jisyo_dirty ){
    /* ѹоݤȤʤʬõ롣*/
    XFillRectangle
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	w->separateWin.roman_ngc,
	x + w->separateWin.roman_width * 11,
	y - w->separateWin.font_ascent,
	w->separateWin.roman_width * 2,
	w->separateWin.font_height ) ;
    /* ⡼ɥ饤ʬ褹롣*/
    skkwin_jputstring
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
	x + w->separateWin.roman_width * 11, y,
	modelineStrings2[ w->separateWin.jisyo_dirty ],
	w->separateWin.roman_width,
	w->separateWin.kanji_width ) ;
    w->separateWin.prev_jisyo_dirty = w->separateWin.jisyo_dirty ;
  }
  if( w->separateWin.prev_chat_adapter != w->separateWin.chat_adapter ){
    /* ѹоݤȤʤʬõ롣*/
    XFillRectangle
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
       w->separateWin.roman_ngc,
       x + w->separateWin.roman_width * 35,
       y - w->separateWin.font_ascent,
       w->separateWin.roman_width * 12,
       w->separateWin.font_height ) ;
    /* ⡼ɥ饤ʬ褹롣*/
    skkwin_jputstring
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
	x + w->separateWin.roman_width * 35, y,
	chatAdapterStrings[ w->separateWin.chat_adapter ],
	w->separateWin.roman_width,
	w->separateWin.kanji_width ) ;
    w->separateWin.prev_chat_adapter = w->separateWin.chat_adapter ;
  }
  if( w->separateWin.prev_eggnl != w->separateWin.egg_like_newline ){
    /* ѹоݤȤʤʬõ롣*/
    XFillRectangle
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
       w->separateWin.roman_ngc,
       x + w->separateWin.roman_width * 47,
       y - w->separateWin.font_ascent,
       w->separateWin.roman_width * 13,
       w->separateWin.font_height ) ;
    /* ⡼ɥ饤ʬ褹롣*/
    if( w->separateWin.egg_like_newline ){
      skkwin_jputstring
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	  w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
	  x + w->separateWin.roman_width * 47, y, " egg-nl)--All",
	  w->separateWin.roman_width,
	  w->separateWin.kanji_width ) ;
    } else {
      skkwin_jputstring
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	  w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
	  x + w->separateWin.roman_width * 47, y, ")--All-------",
	  w->separateWin.roman_width,
	  w->separateWin.kanji_width ) ;
    }
    w->separateWin.prev_eggnl = w->separateWin.egg_like_newline ;
  }
  return ;
}

/*
 * skkinput β̤ʬɽѤɽؿ
 *-----
 * ̾β̤ν(ѹ)ϤδؿˤäƹԤ롣
 */
static void skkinput_DrawScreen( Widget gw )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int x, y, cur_line, linenum, maxline, spos, sty, edy ;
  Window root ;
  unsigned int width, height, border, depth ;
  char *sptr, *dptr ;
  struct SKKInputNode *node = w->separateWin.topbuffer, *mNode ;

  /* 뤬¸ߤƤꡣ*/

  /* Window  size 롣*/
  XGetGeometry( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	        &root, &x, &y, &width, &height, &border, &depth ) ;

  /* ɽǽԿ롣*/
  maxline = height / w->separateWin.font_height ;
  /* ߡƥȤϲԤޤɽʤФʤʤΤ򻻽Ф롣*/
  linenum = skkinput_GetLineNum
    ( node->textbuffer,
      w->separateWin.roman_width, w->separateWin.kanji_width,
      width, node->cur_pos, &cur_line ) ;

  /* ⡼ɥ饤 minibuffer ΤɬפǤ뤫롣*/
  maxline = ( maxline < 3 )? 1 : maxline - 2 ;

#ifdef DEBUG_LV1
  fprintf( stderr, "cursor_line = %d\n", cur_line ) ;
#endif
  /* ɽǽԿĶƤν֤ϰϤ *
   * ǰֺǽιԤֹ椬Ǥ⾮ʤ褦ꤹ롣*/
  if( linenum >= maxline && cur_line >= maxline ){
    int startline = cur_line - maxline + 1 ;
    /* ꤵ줿꤫ɽʤФʤʤΤǡΰ֤򻻽Ф롣*/
    spos = skkinput_FindStartLine
      ( node->textbuffer, w->separateWin.roman_width, 
	w->separateWin.kanji_width, startline, width ) ;
    dptr = node->textbuffer + spos ;
    sptr = w->separateWin.pwrite_string + node->prev_start_pos ;
  }  else {
    spos = 0 ;
    /* ˰ֺǽʸɽǽǤ롣*/
    dptr = node->textbuffer ;
    sptr = w->separateWin.pwrite_string + node->prev_start_pos ;
  }    
  sty = w->separateWin.font_ascent ;
  edy = sty + maxline * w->separateWin.font_height ;
  /* Window  buffer ʸɽ롣*/
  skkinput_DiffDrawWindow
    ( gw, node, dptr, sptr, spos, 0, sty, width, edy ) ;
  strcpy( w->separateWin.pwrite_string, node->textbuffer ) ;
  node->prev_start_pos = spos ;

  /* ⡼ɥ饤ɽ롣ʬɽľбͽǤϤ롣*/
  sty = edy ;
  edy = sty + w->separateWin.font_height ;
  skkinput_DiffDrawModeLine( gw, node, 0, sty, width ) ;
			 
  /* mini-buffer ɽ롣*/
  mNode = w->separateWin.lastbuffer ;

  sty = edy ;
  edy = sty + w->separateWin.font_height ;

  /* for debugging */ 
  if( mNode == node ){
    skkwin_jdiffputstring
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	0, sty, width,
	node->mtextbuffer, w->separateWin.pmtextbuffer,
	w->separateWin.roman_width,
	w->separateWin.kanji_width,
	w->separateWin.font_height,
	w->separateWin.font_ascent ) ;
    strcpy( w->separateWin.pmtextbuffer, node->mtextbuffer ) ;
  } else {
    if( mNode->mtextbuffer[ 0 ] != '\0' ){
      skkwin_jdiffputstring
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	  w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	  w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	  0, sty, width,
	  mNode->mtextbuffer, w->separateWin.pmtextbuffer,
	  w->separateWin.roman_width,
	  w->separateWin.kanji_width,
	  w->separateWin.font_height,
	  w->separateWin.font_ascent ) ;
      strcpy( w->separateWin.pmtextbuffer, mNode->mtextbuffer ) ;
    } else {
      linenum = skkinput_GetLineNum
	( mNode->textbuffer,
	  w->separateWin.roman_width, w->separateWin.kanji_width,
	  width, mNode->cur_pos, &cur_line ) ;
      spos = skkinput_FindStartLine
	( mNode->textbuffer, w->separateWin.roman_width, 
	  w->separateWin.kanji_width, cur_line, width ) ;
      /* Window  buffer ʸɽ롣*/
      skkinput_DiffDrawWindow
	( gw, mNode, mNode->textbuffer + spos,
	  w->separateWin.pmtextbuffer,
	  spos, 0, sty, width, edy ) ;
      strcpy
	( w->separateWin.pmtextbuffer, mNode->textbuffer + spos ) ;
    }
  }
  return ;
}

/*
 * skkinput ɽؿ
 *-----
 * ϺʬԤʤ
 */
static void skkinput_FullDrawWindow
( Widget gw, struct SKKInputNode *node, char *text, 
  int start_pos, int x, int y, int end_x, int end_y )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int pos, chara, kana_flag, kanji_check, cursor_flag ;
  int cx, cy ;

  pos         = start_pos ;
  kana_flag   = False ;
  cursor_flag = False ;

  chara = cx = cy = 0 ;

  while( *text != '\0' ){
    /* ɤȽԤ*/
    kanji_check = 0x80 & *text ;
    /* ɽʤΤɤȽꤹ롣*/
    if( pos == node->cur_pos ){
      if( node->cur_exist ){
	skkinputDrawCursor
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_ngc, w->separateWin.roman_rgc,
	    x, y, w->separateWin.font_ascent,
	    ( kanji_check )? w->separateWin.kanji_width :
	    w->separateWin.roman_width,
	    w->separateWin.font_height,
	    w->separateWin.is_focus ) ;
	cursor_flag = True ;
      }
      cx = x ;
      cy = y ;
    }
    /* ɤΰХܤʤΤǤ礦 */
    if( kanji_check && kana_flag == False ){
      kana_flag = True ;
      chara = ( unsigned char )*text ;
      /* ̤αüޤ褷ΤɤȽǤ롣*/
      if( ( x + w->separateWin.kanji_width ) >= end_x ){
	skkwin_jputchar
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	    x, y, '\\', False ) ;
	skkwin_jputchar
	  ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	    w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	    x + w->separateWin.roman_width, y, '\\', False ) ;
	x = 0 ;
	y += w->separateWin.font_height ;
	/* ̤βޤǽڤäΤɤȽǤ롣*/
	if( y >= end_y )
	  goto exit_loop ;
      }
    } else {
      /* ɤХܤʤΤǤ礦 */
      if( kana_flag ){
	chara = ( chara << 8 ) | ( ( unsigned char )*text ) ;
	/* ξ˥ɽƤΤǤ */
	if( cursor_flag ){
	  skkwin_jputchar
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	      w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	      x, y, chara, False ) ;
	  cursor_flag = False ;
	} else {
	  skkwin_jputchar
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	      w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	      x, y, chara, True ) ;
	}
	x += w->separateWin.kanji_width ;
	kana_flag = False ;
      } else {
	/* ɤäνǤ*/
	if( ( x + w->separateWin.roman_width ) >= end_x ){
	  skkwin_jputchar
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	      x, y, '\\', False ) ;
	  x =  0 ;
	  y += w->separateWin.font_height ;
	  /* ̤βޤǽڤäΤɤȽǤ롣*/
	  if( y >= end_y )
	    goto exit_loop ;
	}
	/* ΰ֤˥뤬ɽƤΤǤСʸȿž *
	 * 롣*/
	if( cursor_flag ){
	  skkwin_jputchar
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_rgc, w->separateWin.kanji_rgc,
	      x, y, *text, False ) ;
	  cursor_flag = False ;
	} else {
	  skkwin_jputchar
	    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	      w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	      x, y, *text, True ) ;
	}
	x += w->separateWin.roman_width ;
	/* ɽʸԥɤäν*/
	if( *text == '\n' ){
	  x =  0 ;
	  y += w->separateWin.font_height ;
	  /* ̤βޤǽڤäΤɤȽǤ롣*/
	  if( y >= end_y )
	    goto exit_loop ;
	}
      }
    }
    text ++ ;
    pos ++ ;
  }
  /* Ǹޤǥ뤬ɽƤʤäΤʤ顢ǸιԤ˥ *
   * ɽ롣*/
  if( pos == node->cur_pos && node->cur_exist ){
    /* Ԥɬפꡩ */
    if( ( x + w->separateWin.roman_width ) > end_x ){
      skkwin_jputchar
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	  w->separateWin.roman_ngc, w->separateWin.kanji_ngc,
	  x, y, '\\', False ) ;
      x  = 0 ;
      y += w->separateWin.font_height ;
    }
    skkinputDrawCursor
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	w->separateWin.roman_ngc, w->separateWin.roman_rgc,
	x, y, w->separateWin.font_ascent, 
	w->separateWin.roman_width, w->separateWin.font_height,
	w->separateWin.is_focus ) ;
    cx = x ;
    cy = y ;
  }

exit_loop:
  node->prev_cx = cx ; 
  node->prev_cy = cy ;
}

static void skkinput_FullDrawModeLine
( Widget gw, struct SKKInputNode *node, int x, int y, unsigned int width )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int modeline = skkinput_GetModelineNumber( node ) ;
  static char *modelineStrings[] = {
    "--SKK:E.:--", "--E.:--", "--E.:--", "--E.:--",
  } ;
  /* νˤäư㤦⡼ɥ饤*/
  static char *modelineStrings2[] = {
    "---Skkinput:", "**-Skkinput:",
  } ;
  static char *chatadapterString[] = {
    "(Fundamental", "(ChatAdapter",
  } ;
  /* ޤ⡼ɥ饤ɽƤõ롣*/
  XFillRectangle
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
      w->separateWin.roman_ngc,
      x, y - w->separateWin.font_ascent,
      width, w->separateWin.font_height ) ;
  /* ⡼ɥ饤ɽ( SKKΥ⡼ɽ )*/
  skkwin_jputstring
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
      w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
      x, y, modelineStrings[ modeline - 1 ],
      w->separateWin.roman_width,
      w->separateWin.kanji_width ) ;
  /* ֤ι*/
  x += w->separateWin.roman_width * 11 ;
  /* ⡼ɥ饤ɽ( ξɽ )*/
  skkwin_jputstring
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	w->separateWin.roman_rgc, w->separateWin.kanji_rgc, 
	x, y, modelineStrings2[ w->separateWin.jisyo_dirty ],
	w->separateWin.roman_width,
	w->separateWin.kanji_width ) ;
  /* ֤ι*/
  x += w->separateWin.roman_width * 24 ;
  /* ᡼⡼ɤɽ*/
  XDrawString
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
      w->separateWin.roman_rgc,
      x, y, chatadapterString[ w->separateWin.chat_adapter % 2 ], 12 ) ;
  /* ֤ι*/
  x += w->separateWin.roman_width * 12 ;
  /* ޥʡ⡼ɤɽ(...ʣä餳դƿϤʥɤ *
   * 줽Ǥ) *
   *     ֤ʤǤ
   * ե꡼֤󤿾ʬäƤΡ
   *     ֿޤ󡣷ʿ˸ȤƤäˤĤƤä
   *           ޤǤ*/
  if( w->separateWin.egg_like_newline ){
    XDrawString
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	w->separateWin.roman_rgc,
	x, y, " egg-nl", 7 ) ;
    x += w->separateWin.roman_width * 7 ;
  } ;
  XDrawString
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
      w->separateWin.roman_rgc,
      x, y, ")", 1 ) ;
  x += w->separateWin.roman_width ;
  /* ³*/
  XDrawString( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	       w->separateWin.roman_rgc,
	       x, y, "--All", 5 ) ;
  x += w->separateWin.roman_width * 5 ;
  for( ; x < width ; x += w->separateWin.roman_width )
    XDrawString
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
	w->separateWin.roman_rgc, x, y, "-", 1 ) ;
  /* ɽ⡼ɥ饤򵭲Ƥ*/
  w->separateWin.prev_modeline     = modeline ;
  w->separateWin.prev_jisyo_dirty  = w->separateWin.jisyo_dirty ;
  w->separateWin.prev_eggnl        = w->separateWin.egg_like_newline ;
  w->separateWin.prev_chat_adapter = w->separateWin.chat_adapter ;
  return ;
}

/*
 * ߥ˥ХåեɽƤΤõؿ
 */
static void SEPW_ClearMinibuffer( Widget gw )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  Window root ;
  int x, y, maxline ;
  unsigned int width, height, border, depth ;

  /* Window  size 롣*/
  XGetGeometry
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
     &root, &x, &y, &width, &height, &border, &depth ) ;

  /* ɽǽԿ롣*/
  maxline = height / w->separateWin.font_height ;
  if( maxline < 3 ){
    y = 2 * w->separateWin.font_height ;
  } else {
    y = ( maxline - 1 ) * w->separateWin.font_height ;
  }
  w->separateWin.pmtextbuffer[ 0 ] = '\0' ;
  XFillRectangle
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
      w->separateWin.roman_rgc,
      0, y, width, w->separateWin.font_height ) ;
  return ;
}

/*
 * skkinput β̤򤤤ľؿ
 *-----
 * ̤礭ѹʤɤԤ줿ˤϤδؿƤФ뤳Ȥ
 * ʤ롣
 */
static void SEPW_RedrawScreen( Widget gw )
{
  SeparateWinWidget w = ( SeparateWinWidget )gw ;
  int x, y, cur_line, linenum, maxline, spos, sty, edy ;
  Window root ;
  unsigned int width, height, border, depth ;
  char *ptr ;
  struct SKKInputNode *mNode, *node = w->separateWin.topbuffer ;

  /* Window Ȥõ롣*/
  XClearWindow( XtDisplay( gw ), XtWindow( w->separateWin.canvas ) ) ;

  /* Window  size 롣*/
  XGetGeometry
    ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ),
     &root, &x, &y, &width, &height, &border, &depth ) ;

  /* ɽǽԿ롣*/
  maxline = height / w->separateWin.font_height ;
  /* ߡƥȤϲԤޤɽʤФʤʤΤ򻻽Ф롣*/
  linenum = skkinput_GetLineNum
    ( node->textbuffer, 
      w->separateWin.roman_width, w->separateWin.kanji_width,
      width, node->cur_pos, &cur_line ) ;
  /* ⡼ɥ饤 minibuffer ΤɬפǤ뤫롣*/
  maxline = ( maxline < 3 )? 1 : maxline - 2 ;
  /* ɽǽԿĶƤν֤ϰϤ *
   * ǰֺǽιԤֹ椬Ǥ⾮ʤ褦ꤹ롣*/
  if( linenum >= maxline && cur_line >= maxline ){
    int startline = cur_line - maxline + 1 ;
    /* ꤵ줿꤫ɽʤФʤʤΤǡΰ֤򻻽Ф롣*/
    spos = skkinput_FindStartLine
      ( node->textbuffer, w->separateWin.roman_width,
	w->separateWin.kanji_width, startline, width ) ;
    ptr = node->textbuffer + spos ;
  }  else {
    spos = 0 ;
    /* ˰ֺǽʸɽǽǤ롣*/
    ptr = node->textbuffer ;
  }    
  sty = w->separateWin.font_ascent ;
  edy = sty + maxline * w->separateWin.font_height ;
  /* Window  buffer ʸɽ롣*/
  skkinput_FullDrawWindow( gw, node, ptr, spos, 0, sty, width, edy ) ;
  strcpy( w->separateWin.pwrite_string, node->textbuffer ) ;

  /* ⡼ɥ饤ɽ롣ʬɽľбͽǤϤ롣*/
  sty = edy ;
  edy = sty + w->separateWin.font_height ;
  skkinput_FullDrawModeLine( gw, node, 0, sty, width ) ;
			 
  /* mini-buffer ɽ롣*/
  mNode = w->separateWin.lastbuffer ;
  sty = edy ;
  edy = sty + w->separateWin.font_height ;

  if( mNode == node ){
    skkwin_jputstring
      ( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	w->separateWin.roman_ngc, w->separateWin.kanji_ngc, 
	0, sty, mNode->mtextbuffer,
	w->separateWin.roman_width,
	w->separateWin.kanji_width ) ;
    strcpy( w->separateWin.pmtextbuffer, node->mtextbuffer ) ;
  } else {
    if( mNode->mtextbuffer[ 0 ] != '\0' ){
      skkwin_jputstring
	( XtDisplay( gw ), XtWindow( w->separateWin.canvas ), 
	  w->separateWin.roman_ngc, w->separateWin.kanji_ngc, 
	  0, sty, mNode->mtextbuffer,
	  w->separateWin.roman_width,
	  w->separateWin.kanji_width ) ;
      strcpy( w->separateWin.pmtextbuffer, mNode->mtextbuffer ) ;
    } else {
      linenum = skkinput_GetLineNum
	( mNode->textbuffer, 
	  w->separateWin.roman_width, w->separateWin.kanji_width,
	  width, mNode->cur_pos, &cur_line ) ;
      /* ꤵ줿꤫ɽʤФʤʤΤǡΰ֤򻻽Ф롣*/
      spos = skkinput_FindStartLine
	( mNode->textbuffer, w->separateWin.roman_width,
	  w->separateWin.kanji_width, cur_line, width ) ;
      /* Window  buffer ʸɽ롣*/
      skkinput_FullDrawWindow
	( gw, mNode, mNode->textbuffer + spos,
	  spos, 0, sty, width, edy ) ;
      w->separateWin.pmtextbuffer[ 0 ] = '\0' ;
    }
  }
  return ;
}

/*
 * ʬΥκԤ˸ƤӽФ륳Хåؿ
 */
static void SeparateCanvas_RedisplayCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  SEPW_RedrawScreen( ( Widget )client ) ;
  return ;
}

/*
 * Canvas Widget ˸ƤӽФ륳Хåؿ
 */
static void Canvas_DestroyCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  XtRemoveAllCallbacks( gw, XtNredrawNotify ) ;
  XtRemoveAllCallbacks( gw, XtNdestroyNotify ) ;
  return ;
}

static void WMMessageCloseCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  /* ƤΥХåؿ˴Ƥ*/
  XtRemoveAllCallbacks( gw, XtNwmcloseNotify ) ;
  XtRemoveAllCallbacks( gw, XtNdestroyNotify ) ;
  XtDestroyWidget( ( Widget )client ) ;
  return ;
}

static void WMShellDestroyCallback
( Widget gw, caddr_t client, caddr_t caller )
{
  /* ƤΥХåؿ˴Ƥ*/
  XtRemoveAllCallbacks( gw, XtNwmcloseNotify ) ;
  XtRemoveAllCallbacks( gw, XtNdestroyNotify ) ;
  return ;
}
